"""Test kugar parser module."""

import unittest
from pineboolib.application.parsers.parser_kut.tests import fixture_path
from pineboolib.loader.main import init_testing, finish_testing
from pineboolib import application
from pineboolib.core.utils import utils_base


class TestParser(unittest.TestCase):
    """Test Parsing KUT to PDF."""

    @classmethod
    def setUpClass(cls) -> None:
        """Init test project."""
        utils_base.FORCE_DESKTOP = True
        init_testing()

    def test_kugar_parser_1(self) -> None:
        """Test parser."""

        from pineboolib.qsa import qsa

        from pineboolib.plugins.mainform.eneboo import eneboo
        import os

        main_form_class = getattr(eneboo, "MainForm", None)
        self.assertTrue(main_form_class)
        application.PROJECT.main_window = main_form_class()  # type: ignore[misc]
        self.assertTrue(application.PROJECT.main_window)
        if application.PROJECT.main_window is not None:
            application.PROJECT.main_window.initScript()

        qsa_sys = qsa.sys
        path = fixture_path("principal.eneboopkg")
        self.assertTrue(os.path.exists(path))
        self.assertTrue(qsa_sys.loadModules(path, False))

        # qsa.from_project("flfactppal").iface.valoresIniciales()
        cur_paises = qsa.FLSqlCursor("paises")
        cur_paises.setModeAccess(cur_paises.Insert)
        cur_paises.refreshBuffer()
        cur_paises.setValueBuffer("codpais", "ES")
        cur_paises.setValueBuffer("nombre", "ESPAÑA")
        self.assertTrue(cur_paises.commitBuffer())
        cur_paises.setModeAccess(cur_paises.Insert)
        cur_paises.refreshBuffer()
        cur_paises.setValueBuffer("codpais", "PT")
        cur_paises.setValueBuffer("nombre", "PORTUGAL")
        self.assertTrue(cur_paises.commitBuffer())

        cur_paises.select("1=1")
        cur_paises.first()
        init_ = cur_paises.valueBuffer("codpais")
        self.assertEqual(cur_paises.valueBuffer("nombre"), "ESPAÑA")
        cur_paises.last()
        last_ = cur_paises.valueBuffer("codpais")
        qry_paises = qsa.FLSqlQuery("paises")
        qry_paises.setValueParam("from", init_)
        qry_paises.setValueParam("to", last_)

        rpt_viewer_ = qsa.FLReportViewer()
        rpt_viewer_.setReportTemplate("paises")
        rpt_viewer_.setReportData(qry_paises)

        rpt_viewer_.renderReport()
        pdf_file = None
        if rpt_viewer_._report_engine and hasattr(rpt_viewer_._report_engine, "_parser"):
            pdf_file = rpt_viewer_._report_engine._parser.get_file_name()

        self.assertTrue(pdf_file)

    def test_parser_tools_1(self) -> None:
        """Test parser tools."""
        from pineboolib.application.parsers.parser_kut import kparsertools
        from xml.etree import ElementTree as et
        from pineboolib.core.utils.utils_base import load2xml
        from pineboolib.application.database import pnsqlquery, pnsqlcursor
        from pineboolib.qsa import qsa
        import datetime
        import os

        qry = pnsqlquery.PNSqlQuery()
        qry.setTablesList("paises")
        qry.setSelect("codpais, bandera")
        qry.setFrom("paises")
        qry.setWhere("1=1")
        self.assertTrue(qry.exec_())
        self.assertTrue(qry.first())
        data = qsa.sys.toXmlReportData(qry)
        parser_tools = kparsertools.KParserTools()
        xml_data = load2xml(data.toString()).getroot()

        child = xml_data.findall("Row")[0]
        element = parser_tools.convertToNode(child)
        self.assertTrue(element)
        fecha_ = str(datetime.date.__format__(datetime.date.today(), "%d-%m-%Y"))

        self.assertEqual(parser_tools.getSpecial("Fecha"), fecha_)
        self.assertEqual(parser_tools.getSpecial("[Date]"), fecha_)
        self.assertEqual(parser_tools.getSpecial("NúmPágina", 1), "1")
        self.assertEqual(parser_tools.getSpecial("PageNo", 6), "6")
        self.assertEqual(parser_tools.getSpecial("[NÃºmPÃ¡gina]", 12), "12")
        from PyQt6 import QtCore  # type: ignore[import]

        ret_ = QtCore.QLocale.system().toString(float("11.22"), "f", 2)
        xml = et.Element("data")
        item_1 = et.SubElement(xml, "Item")
        item_1.set("Precision", "2")
        self.assertEqual(parser_tools.calculated("11.22", 2, item_1), ret_)
        self.assertEqual(parser_tools.calculated("2019-01-31T00:01:02", 3), "31-01-2019")
        self.assertEqual(parser_tools.calculated("codpais", 1, None, child), "ES")  # type: ignore [arg-type]

        cur = pnsqlcursor.PNSqlCursor("paises")
        cur.select("1=1")
        cur.first()
        buffer = cur.buffer()
        if buffer:
            bandera = buffer.value("bandera")
            self.assertEqual(
                parser_tools.parseKey(str(bandera)),
                os.path.abspath("%s/%s.png" % (application.PROJECT.tmpdir, bandera)),
            )

    def test_parser_tools_2(self) -> None:
        """Test parser tools."""
        from pineboolib.application.parsers.parser_kut import kparsertools
        from xml.etree import ElementTree as et
        from decimal import Decimal

        parser_tools = kparsertools.KParserTools()
        self.assertEqual(parser_tools.convertPageSize(0, 0), [595, 842])
        self.assertEqual(parser_tools.convertPageSize(1, 0), [709, 499])
        self.assertEqual(parser_tools.convertPageSize(2, 0), [612, 791])
        self.assertEqual(parser_tools.convertPageSize(3, 0), [612, 1009])
        self.assertEqual(parser_tools.convertPageSize(5, 0), [2384, 3370])
        self.assertEqual(parser_tools.convertPageSize(6, 0), [1684, 2384])
        self.assertEqual(parser_tools.convertPageSize(7, 0), [1191, 1684])
        self.assertEqual(parser_tools.convertPageSize(8, 0), [842, 1191])
        self.assertEqual(parser_tools.convertPageSize(9, 0), [420, 595])
        self.assertEqual(parser_tools.convertPageSize(10, 0), [298, 420])
        self.assertEqual(parser_tools.convertPageSize(11, 0), [210, 298])
        self.assertEqual(parser_tools.convertPageSize(12, 0), [147, 210])
        self.assertEqual(parser_tools.convertPageSize(13, 0), [105, 147])
        self.assertEqual(parser_tools.convertPageSize(14, 0), [4008, 2835])
        self.assertEqual(parser_tools.convertPageSize(15, 0), [2835, 2004])
        self.assertEqual(parser_tools.convertPageSize(16, 0), [125, 88])
        self.assertEqual(parser_tools.convertPageSize(17, 0), [2004, 1417])
        self.assertEqual(parser_tools.convertPageSize(18, 0), [1417, 1001])
        self.assertEqual(parser_tools.convertPageSize(19, 0), [1001, 709])
        self.assertEqual(parser_tools.convertPageSize(20, 0), [499, 354])
        self.assertEqual(parser_tools.convertPageSize(21, 0), [324, 249])
        self.assertEqual(parser_tools.convertPageSize(22, 0), [249, 176])
        self.assertEqual(parser_tools.convertPageSize(23, 0), [176, 125])
        self.assertEqual(parser_tools.convertPageSize(24, 0), [649, 459])
        self.assertEqual(parser_tools.convertPageSize(25, 0), [113, 79])
        self.assertEqual(parser_tools.convertPageSize(28, 0), [1255, 791])
        self.assertEqual(parser_tools.convertPageSize(29, 0), [791, 1255])
        self.assertEqual(parser_tools.convertPageSize(30, 0, [100, 200]), [100, 200])
        self.assertEqual(parser_tools.convertPageSize(100, 0), [595, 842])

        xml = et.Element("AllItems")
        item_1 = et.SubElement(xml, "Item")
        item_1.set("valor", "16.01")
        item_1.set("level", "0")
        ret_0 = float(parser_tools.calculate_sum("valor", item_1, xml.findall("Item"), 0))  # 16.01
        item_2 = et.SubElement(xml, "Item")
        item_2.set("valor", "5.84")
        item_2.set("level", "1")
        ret_1 = float(parser_tools.calculate_sum("valor", item_2, xml.findall("Item"), 1))  # 5.84
        item_4 = et.SubElement(xml, "Item")
        item_4.set("valor", "11.00")
        item_4.set("level", "0")
        ret_4 = float(parser_tools.calculate_sum("valor", item_4, xml.findall("Item"), 0))  # 11
        item_3 = et.SubElement(xml, "Item")
        item_3.set("valor", "26.29")
        item_3.set("level", "1")
        ret_3 = float(parser_tools.calculate_sum("valor", item_3, xml.findall("Item"), 1))  # 26.29
        item_6 = et.SubElement(xml, "Item")
        item_6.set("valor", "26.29")
        item_6.set("level", "1")

        ret_5 = float(
            parser_tools.calculate_sum("valor", item_6, xml.findall("Item"), 1)
        )  # 26.29 + 26.29

        item_5 = et.SubElement(xml, "Item")
        item_5.set("valor", "6.29")
        item_5.set("level", "2")

        ret_2 = float(parser_tools.calculate_sum("valor", item_5, xml.findall("Item"), 2))  # 6.29

        item_7 = et.SubElement(xml, "Item")
        item_7.set("valor", "1.01")
        item_7.set("level", "0")

        item_9 = et.SubElement(xml, "Item")
        item_9.set("valor", "0")
        item_9.set("level", "1")

        item_8 = et.SubElement(xml, "Item")
        item_8.set("valor", "2")
        item_8.set("level", "2")

        item_10 = et.SubElement(xml, "Item")
        item_10.set("valor", "1")
        item_10.set("level", "2")

        ret_6 = float(parser_tools.calculate_sum("valor", item_8, xml.findall("Item"), 0))  # 2

        ret_7 = float(parser_tools.calculate_sum("valor", item_7, xml.findall("Item"), 1))  # 0

        ret_8 = float(parser_tools.calculate_sum("valor", item_8, xml.findall("Item"), 2))  # 2
        ret_9 = float(parser_tools.calculate_sum("valor", item_10, xml.findall("Item"), 2))  # 3
        ret_10 = float(parser_tools.calculate_sum("valor", item_10, xml.findall("Item"), 0))  # 3

        self.assertEqual(float(Decimal(format(ret_0, ".2f"))), 16.01)  # 16.01
        self.assertEqual(float(Decimal(format(ret_1, ".2f"))), 5.84)  # 5.84
        self.assertEqual(float(Decimal(format(ret_3, ".2f"))), 26.29)  # 26.29
        self.assertEqual(float(Decimal(format(ret_5, ".2f"))), 52.58)  # 26.29 + 26.29
        self.assertEqual(float(Decimal(format(ret_4, ".2f"))), 11)  # 11
        self.assertEqual(float(Decimal(format(ret_2, ".2f"))), 6.29)  # 6.29
        self.assertEqual(float(Decimal(format(ret_6, ".2f"))), 2)  # 2
        self.assertEqual(float(Decimal(format(ret_7, ".2f"))), 0)  # 0
        self.assertEqual(float(Decimal(format(ret_8, ".2f"))), 2)  # 2
        self.assertEqual(float(Decimal(format(ret_9, ".2f"))), 3)  # 3
        self.assertEqual(float(Decimal(format(ret_10, ".2f"))), 3)  # 3

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()



# lextab.py. This file automatically created by PLY (version 3.11). Don't edit!
_tabversion   = '3.10'
_lextokens    = set(('AND', 'AT', 'BACKSLASH', 'BREAK', 'CASE', 'CATCH', 'CCONST', 'CLASS', 'COLON', 'COMMA', 'COMMENTCLOSE', 'CONDITIONAL1', 'CONST', 'CONTINUE', 'DEFAULT', 'DELETE', 'DIVEQUAL', 'DIVIDE', 'DO', 'DOCSTRINGOPEN', 'DOLLAR', 'DQOUTE', 'ELSE', 'EQ', 'EQQ', 'EQUALS', 'EXTENDS', 'FCONST', 'FOR', 'FUNCTION', 'GE', 'GT', 'ICONST', 'ID', 'IF', 'IN', 'LAND', 'LBRACE', 'LBRACKET', 'LE', 'LNOT', 'LOR', 'LPAREN', 'LSHIFT', 'LT', 'MINUS', 'MINUSEQUAL', 'MINUSMINUS', 'MOD', 'MODEQUAL', 'NE', 'NEQ', 'NEW', 'OR', 'PERIOD', 'PLUS', 'PLUSEQUAL', 'PLUSPLUS', 'RBRACE', 'RBRACKET', 'RETURN', 'RPAREN', 'RSHIFT', 'SCONST', 'SEMI', 'SQOUTE', 'STATIC', 'SWITCH', 'THROW', 'TIMES', 'TIMESEQUAL', 'TRY', 'TYPEOF', 'VAR', 'WHILE', 'WITH', 'XOR'))
_lexreflags   = 64
_lexliterals  = ''
_lexstateinfo = {'INITIAL': 'inclusive'}
_lexstatere   = {'INITIAL': [('(?P<t_NEWLINE>\\n+)|(?P<t_ID>[A-Za-z_]+[\\w_]*)|(?P<t_comment>(/\\*( |\\*\\*)(.|\\n)*?\\*/)|(//.*))|(?P<t_DOCSTRINGOPEN>/\\*\\*[ ]+)|(?P<t_preprocessor>\\#(.)*?\\n)|(?P<t_FCONST>((\\d+)(\\.\\d+)(e(\\+|-)?(\\d+))? | (\\d+)e(\\+|-)?(\\d+))([lL]|[fF])?)|(?P<t_ICONST>\\d+([uU]|[lL]|[uU][lL]|[lL][uU])?)|(?P<t_CCONST>\\\'([^\\\'\\\\\\n]|(\\\\.)|\\\\\\n)*?\\\')|(?P<t_SCONST>\\"([^\\"\\\\\\n]|(\\\\.)|\\\\\\n)*?\\")|(?P<t_LOR>\\|\\|)|(?P<t_PLUSPLUS>\\+\\+)|(?P<t_COMMENTCLOSE>\\*/)|(?P<t_EQQ>===)|(?P<t_NEQ>!==)|(?P<t_PLUSEQUAL>\\+=)|(?P<t_TIMESEQUAL>\\*=)|(?P<t_BACKSLASH>\\\\)|(?P<t_CONDITIONAL1>\\?)|(?P<t_DIVEQUAL>/=)|(?P<t_DOLLAR>\\$)|(?P<t_EQ>==)|(?P<t_GE>>=)|(?P<t_LAND>&&)|(?P<t_LBRACE>\\{)|(?P<t_LBRACKET>\\[)|(?P<t_LE><=)|(?P<t_LPAREN>\\()|(?P<t_LSHIFT><<)|(?P<t_MINUSEQUAL>-=)|(?P<t_MINUSMINUS>--)|(?P<t_MODEQUAL>%=)|(?P<t_NE>!=)|(?P<t_OR>\\|)|(?P<t_PERIOD>\\.)|(?P<t_PLUS>\\+)|(?P<t_RBRACE>\\})|(?P<t_RBRACKET>\\])|(?P<t_RPAREN>\\))|(?P<t_RSHIFT>>>)|(?P<t_TIMES>\\*)|(?P<t_XOR>\\^)|(?P<t_AND>&)|(?P<t_AT>@)|(?P<t_COLON>:)|(?P<t_COMMA>,)|(?P<t_DIVIDE>/)|(?P<t_DQOUTE>")|(?P<t_EQUALS>=)|(?P<t_GT>>)|(?P<t_LNOT>!)|(?P<t_LT><)|(?P<t_MINUS>-)|(?P<t_MOD>%)|(?P<t_SEMI>;)|(?P<t_SQOUTE>\')', [None, ('t_NEWLINE', 'NEWLINE'), ('t_ID', 'ID'), ('t_comment', 'comment'), None, None, None, None, ('t_DOCSTRINGOPEN', 'DOCSTRINGOPEN'), ('t_preprocessor', 'preprocessor'), None, (None, 'FCONST'), None, None, None, None, None, None, None, None, None, None, (None, 'ICONST'), None, (None, 'CCONST'), None, None, (None, 'SCONST'), None, None, (None, 'LOR'), (None, 'PLUSPLUS'), (None, 'COMMENTCLOSE'), (None, 'EQQ'), (None, 'NEQ'), (None, 'PLUSEQUAL'), (None, 'TIMESEQUAL'), (None, 'BACKSLASH'), (None, 'CONDITIONAL1'), (None, 'DIVEQUAL'), (None, 'DOLLAR'), (None, 'EQ'), (None, 'GE'), (None, 'LAND'), (None, 'LBRACE'), (None, 'LBRACKET'), (None, 'LE'), (None, 'LPAREN'), (None, 'LSHIFT'), (None, 'MINUSEQUAL'), (None, 'MINUSMINUS'), (None, 'MODEQUAL'), (None, 'NE'), (None, 'OR'), (None, 'PERIOD'), (None, 'PLUS'), (None, 'RBRACE'), (None, 'RBRACKET'), (None, 'RPAREN'), (None, 'RSHIFT'), (None, 'TIMES'), (None, 'XOR'), (None, 'AND'), (None, 'AT'), (None, 'COLON'), (None, 'COMMA'), (None, 'DIVIDE'), (None, 'DQOUTE'), (None, 'EQUALS'), (None, 'GT'), (None, 'LNOT'), (None, 'LT'), (None, 'MINUS'), (None, 'MOD'), (None, 'SEMI'), (None, 'SQOUTE')])]}
_lexstateignore = {'INITIAL': ' \r\t\x0c'}
_lexstateerrorf = {'INITIAL': 't_error'}
_lexstateeoff = {}



"""
Database package to manage data.

This package holds all functions and classes to manage the application data.
"""

from pineboolib.application.database import pnsignals

DB_SIGNALS = pnsignals.PNSignals()



"""Dummy signal module."""

from typing import List, Callable, TYPE_CHECKING
from pineboolib.application import connections

if TYPE_CHECKING:
    from pineboolib.application.database.orm import basemodel


class FakeSignal(object):
    """FakeSignal class."""

    _remote_funcs: List[Callable]
    _parent_model: "basemodel.BaseModel"

    def __init__(self, parent_model: "basemodel.BaseModel"):
        """Initialice."""
        self._parent_model = parent_model
        self._remote_funcs = []

    def connect(self, func_: Callable) -> None:
        """Set a function to connect."""

        if func_ not in self._remote_funcs:
            self._remote_funcs.append(func_)

    def disconnect(self, func_: Callable) -> None:
        """Set a function to disconnect."""

        if func_ in self._remote_funcs:
            self._remote_funcs.remove(func_)

    def emit(self, text: str) -> None:
        """Call all conected functions."""
        for func_ in self._remote_funcs:
            func_(*[text, self._parent_model.cursor][: connections.get_expected_args_num(func_)])



# -*- coding: utf-8 -*-
# Translated with pineboolib v0.71.18
"""Fserial_model module."""

import sqlalchemy  # type: ignore [import] # noqa: F821

from pineboolib.application.database.orm import basemodel


class Flserial(basemodel.BaseModel):  # type: ignore [misc] # noqa: F821
    """Flserial class."""

    __tablename__ = "flserial"

    # --- Metadata --->
    legacy_metadata = {
        "name": "flserial",
        "alias": "Serie",
        "fields": [
            {"name": "serie", "alias": "Serie", "pk": True, "type": "serial", "null": False},
            {
                "name": "sha",
                "alias": "SHA1",
                "type": "string",
                "length": 255,
                "calculated": True,
                "editable": False,
            },
        ],
    }

    # <--- Metadata ---

    # --- Fields --->

    serie = sqlalchemy.Column("serie", sqlalchemy.Integer, primary_key=True)
    sha = sqlalchemy.Column("sha", sqlalchemy.String(255))


# <--- Fields ---



"""Test_qpushbutton module."""

import unittest


class TestQPushButton(unittest.TestCase):
    """TestQPushButton Class."""

    def test_enabled(self) -> None:
        """Test if the control enabled/disabled."""

        from pineboolib.q3widgets import qpushbutton

        button = qpushbutton.QPushButton()
        self.assertTrue(button.enabled)
        button.enabled = False
        self.assertFalse(button.enabled)

    def test_label(self) -> None:
        """Test label."""

        from pineboolib.q3widgets import qpushbutton

        button = qpushbutton.QPushButton()
        button.setTextLabel("etiqueta")
        text: str = str(button.text)  # type: ignore
        self.assertEqual(text, "etiqueta")



"""Qwidget module."""

# -*- coding: utf-8 -*-
from PyQt6 import QtWidgets, QtCore  # type: ignore[import]
from typing import cast, Any
from pineboolib.core import decorators


class QWidget(QtWidgets.QWidget):
    """QWidget class."""

    def child(self, child_name: str) -> QtWidgets.QWidget:
        """Return an QWidget especified by name."""

        ret = cast(QtWidgets.QWidget, self.findChild(QtWidgets.QWidget, child_name))

        return ret or QWidget()

    def get_title(self) -> str:
        """Return widget title."""
        return self.windowTitle()

    def set_title(self, title: str) -> None:
        """Set title."""
        self.setWindowTitle(title)

    @decorators.not_implemented_warn
    def setInsideMargin(self, value):
        """Set inside margin."""
        pass

    @decorators.not_implemented_warn
    def setInsideSpacing(self, value):
        """Set inside margin."""
        pass

    def getattr(self, name: str) -> Any:
        """Return an attribute specified by name."""

        if name == "name":
            return self.objectName()
        else:
            print("FIXME:Q3Widget:", name)
            return getattr(QtCore.Qt, name, None)

    title = property(get_title, set_title)



"""Pdf_qr module."""

from PyQt6 import QtGui, QtCore  # type: ignore[import]

from pineboolib import application
from pineboolib.core.utils import logging

from pdf2image import convert_from_path  # type: ignore[import]
from PIL import Image, ImageQt  # type: ignore[import]
import os
import io
import qrcode  # type: ignore[import]


from typing import List, Optional

LOGGER = logging.get_logger(__name__)


class PdfQr:
    """PdfQr class."""

    _orig: str
    _pos_x: int
    _pos_y: int
    _all_pages: bool
    _size: int
    _text: str
    _qr_text: str
    _signed_data: List["QtGui.QImage"]
    _font_name: str
    _font_size: int
    _show_text: bool
    _dpi: int
    _tmp_qr_img: str
    _factor: float
    _qr_image: Optional["QtGui.QImage"]

    def __init__(self, orig: str = "") -> None:
        """Initialize."""

        self._orig = orig
        self._pos_x = 100
        self._pos_y = 100
        self._all_pages = False
        self._size = 7  # 5
        self._text = ""
        self._qr_text = ""
        self._signed_data = []
        self._font_name = "Arial"
        self._font_size = 8
        self._show_text = False
        self._dpi = 300
        self._tmp_qr_img = ""
        self._factor = self._dpi / 100
        self._qr_image = None
        self._ext = "PNG"

    def set_size(self, size: int) -> None:
        """Set size."""

        # (size * 4 ) + 17

        self._size = int(size)

    def set_extension(self, ext_name: str = "PNG") -> None:
        """Set extension."""

        self._ext = ext_name

    def set_dpi(self, dpi: int = 300) -> None:
        """Set dpi."""

        self._dpi = int(dpi)
        self._factor = self._dpi / 100

    def set_text(self, qr_text: str, text: str = "") -> None:
        """Set text."""

        self._qr_text = qr_text
        if not text:
            self._show_text = False
            text = qr_text
        else:
            self._show_text = True
        self._text = text

    def set_font(self, name: str, size: int) -> None:
        """Set font name and size."""

        self._font_name = name
        self._font_size = int(size)

    def set_position(self, pos_x: int, pos_y: int) -> None:
        """Set Possition. 0,0 = Bottom Right."""
        self._pos_x = int(pos_x)
        self._pos_y = int(pos_y)

    def sign(self, all_pages: bool = False) -> bool:
        """Sing file."""

        if not self._orig:
            LOGGER.warning("A file has not been specified to sign")
            return False

        self._all_pages = all_pages
        if not self.build_qr():
            return False

        signed_image = self._qr_image
        if not signed_image:
            LOGGER.warning("QR Image not found")
            return False

        mark = True
        pos_x = self._pos_x
        pos_y = self._pos_y
        self._signed_data = []
        pages = convert_from_path(
            self._orig, dpi=self._dpi, output_folder=application.PROJECT.tmpdir  # size=500
        )
        for num, page in enumerate(pages):
            page_image = ImageQt.ImageQt(page)
            if mark:
                painter = QtGui.QPainter()
                painter.begin(page_image)
                painter.setCompositionMode(
                    QtGui.QPainter.CompositionMode.CompositionMode_SourceOver
                )
                painter.drawImage(
                    int(page_image.width() - (pos_x * self._factor) - signed_image.width()),
                    int(page_image.height() - (pos_y * self._factor) - signed_image.height()),
                    signed_image,
                )
                painter.end()

            self._signed_data.append(page_image)

            if not self._all_pages:
                mark = False

        return True

    def get_qr(self) -> str:
        """Return QR temp file."""

        if not self._tmp_qr_img:
            if not self.build_qr():
                return ""

        return self._tmp_qr_img

    def build_qr(self) -> bool:
        """Build QR image."""

        try:
            qr_ = qrcode.QRCode(
                version=self._size,
                error_correction=qrcode.constants.ERROR_CORRECT_M,
                box_size=3,
                border=4,
            )  # 135 px x 135 px = (35.89 mm x 35.89 mm)
            qr_.add_data(self._qr_text)
            qr_.make(fit=False)
            qr_image = qr_.make_image(fill_color="black", back_color="white")

            qr_folder = os.path.join(
                application.PROJECT.tmpdir,
                "cache",
                application.PROJECT.conn_manager.mainConn().driver().DBName(),
                "QR",
            )

            if not os.path.exists(qr_folder):
                os.mkdir(qr_folder)

            self._tmp_qr_img = os.path.join(
                qr_folder,
                "%s.%s"
                % (QtCore.QDateTime.currentDateTime().toString("ddMMyyyyhhmmsszzz"), self._ext),
            )
            qr_image.save(self._tmp_qr_img, self._ext)

            if self._show_text:
                image_qr = QtGui.QImage(self._tmp_qr_img)
                image_label = QtGui.QImage(self._tmp_qr_img)
                size_font = int(self._font_size * self._factor)
                text_list = self._text.split("\n")
                max_text_len = 0
                for text in text_list:
                    current_len = len(text)
                    if current_len > max_text_len:
                        max_text_len = current_len

                text_width = max_text_len * (self._font_size - 1.5)
                extra_width = text_width if text_width > (qr_image.height) else qr_image.height
                print(
                    "extra:%s, qr:%s, max:%s, text:%s, size: %s"
                    % (extra_width, image_qr.width(), text_width, len(self._text), self._font_size)
                )
                extra_height = (self._font_size + 2) * len(text_list)

                image_label_resized = image_label.scaled(
                    int(extra_width * self._factor),
                    int((qr_image.height + extra_height) * self._factor),
                )
                image_qr = image_qr.scaled(
                    int(image_qr.width() * self._factor), int(image_qr.height() * self._factor)
                )
                image_label_resized.fill(0)
                label_painter = QtGui.QPainter()
                label_painter.begin(image_label_resized)
                label_painter.setCompositionMode(
                    QtGui.QPainter.CompositionMode.CompositionMode_SourceOver
                )
                label_painter.setPen(QtGui.QPen(QtCore.Qt.GlobalColor.black))
                label_painter.setFont(
                    QtGui.QFont(
                        self._font_name,
                        size_font,
                        QtGui.QFont.Weight.Normal,
                    )
                )

                proccessed_text = ""
                for text in text_list:
                    proccessed_text += "      %s\n" % text

                label_painter.drawText(
                    image_label_resized.rect(),
                    QtCore.Qt.AlignmentFlag.AlignTop,
                    proccessed_text,
                )
                label_painter.setCompositionMode(
                    QtGui.QPainter.CompositionMode.CompositionMode_SourceOver
                )
                label_painter.drawImage(
                    image_label_resized.width() - image_qr.width(),
                    int(((self._font_size + 4) * len(text_list)) * self._factor),
                    image_qr,
                )
                label_painter.end()
                image_label_resized.save(self._tmp_qr_img, "PNG")
                self._qr_image = image_label_resized
            else:
                signed_image = QtGui.QImage(self._tmp_qr_img)
                self._qr_image = signed_image.scaled(
                    int(signed_image.width() * self._factor),
                    int(signed_image.height() * self._factor),
                )

        except Exception as error:
            LOGGER.warning("Build QR failed:%s", str(error))
            return False

        return True

    def save_file(self, file_path: str) -> bool:
        """Save file."""

        if not file_path:
            LOGGER.warning("File path %s is empty!", file_path)
            return False

        if not self._signed_data:
            LOGGER.warning("Data is empty!")
            return False

        first = True
        for img_data in self._signed_data:
            buffer = QtCore.QBuffer()
            buffer.open(QtCore.QIODeviceBase.OpenModeFlag.ReadWrite)
            img_data.save(buffer, "PNG")
            page = Image.open(io.BytesIO(buffer.data()))  # type: ignore[arg-type] # noqa: F821
            page.save(
                file_path,
                resolution=self._dpi,
                append=not first,
                author="Pineboo ERP",
                title="pdf signed",
            )
            first = False

        return True



"""projectconfig module."""

import base64
import hashlib
import os
import pathlib
from cryptography import fernet  # type: ignore [import] # noqa: F821

from xml.etree import ElementTree as ELT
from typing import Tuple, Optional, Any

from pineboolib import logging
from pineboolib.core.utils import utils_base
from pineboolib.core.utils import version as utils_version
from pineboolib.core import settings

VERSION_1_0 = utils_version.VersionNumber("1.0")
VERSION_1_1 = utils_version.VersionNumber("1.1")
VERSION_1_2 = utils_version.VersionNumber("1.2")

LOGGER = logging.get_logger(__name__)


class ProjectConfig:
    """
    Read and write XML on profiles. Represents a database connection configuration.
    """

    SAVE_VERSION = VERSION_1_2  #: Version for saving

    #: Folder where to read/write project configs.
    profile_dir: str = utils_base.filedir(
        settings.CONFIG.value(
            "ebcomportamiento/profiles_folder", "%s/Pineboo/profiles" % pathlib.Path.home()
        )
    )

    version: "utils_version.VersionNumber"  #: Version number for the profile read.
    fernet: Optional["fernet.Fernet"]  #: Cipher used, if any.

    database: str  #: Database Name, file path to store it, or :memory:
    host: Optional[str]  #: DB server Hostname. None for local files.
    port: Optional[int]  #: DB server port. None for local files.
    username: Optional[str]  #: Database User login name.
    password: Optional[str]  #: Database User login password.
    type: str  #: Driver Type name to use when connecting
    project_password: str  #: Password to cipher when load/saving. Empty string for no ciphering.
    password_required: bool  #: True if a password is required to read data. (Was partially loaded.)
    description: str  #: Project name in GUI
    filename: str  #: File path to read / write this project from / to

    def __init__(
        self,
        database: Optional[str] = None,
        host: Optional[str] = None,
        port: Optional[int] = None,
        type: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        load_xml: Optional[str] = None,
        connstring: Optional[str] = None,
        description: Optional[str] = None,
        filename: Optional[str] = None,
        project_password: str = "",
    ) -> None:
        """Initialize."""
        self.project_password = project_password
        self.password_required = False
        self.version = self.SAVE_VERSION
        self.fernet = None

        if connstring:
            username, password, type, host, port, database = self.translate_connstring(connstring)
        elif load_xml:
            self.filename = os.path.join(
                self.profile_dir, load_xml if load_xml.endswith("xml") else "%s.xml" % load_xml
            )
            self.load_projectxml()
            return
        if database is None:
            raise ValueError("Database is mandatory. Or use load_xml / connstring params")
        if type is None:
            raise ValueError("Type is mandatory. Or use load_xml / connstring params")
        self.database = database
        self.host = host
        self.port = port
        self.type = type
        self.username = username
        self.password = password
        self.description = description if description else "unnamed"
        if filename is None:
            file_basename = self.description.lower().replace(" ", "_")
            self.filename = os.path.join(self.profile_dir, "%s.xml" % file_basename)
        else:
            self.filename = os.path.join(self.profile_dir, filename)

    def get_uri(self, show_password: bool = False) -> str:
        """Get connection as an URI."""
        host_port = ""
        if self.host:
            host_port += self.host
        if self.port:
            host_port += ":%d" % self.port

        user_pass = ""
        if self.username:
            user_pass += self.username
        if self.password:
            if show_password:
                user_pass += ":%s" % self.password
            else:
                pass_bytes: bytes = hashlib.sha256(self.password.encode()).digest()
                user_pass += ":*" + base64.b64encode(pass_bytes).decode()[:4]

        if user_pass:
            user_pass = "@%s" % user_pass

        uri = host_port + user_pass

        if self.database:
            if uri:
                uri += "/"
            uri += self.database

        return "[%s]://%s" % (self.type, uri)

    def __repr__(self) -> str:
        """Display the information in text mode."""
        if self.project_password:
            # 4 chars in base-64 is 3 bytes. 256**3 should be enough to know if you have the wrong
            # password.
            pass_bytes: bytes = hashlib.sha256(self.project_password.encode()).digest()
            passwd = "-" + base64.b64encode(pass_bytes).decode()[:4]
        else:
            passwd = ""
        return "<ProjectConfig%s name=%r uri=%r>" % (
            passwd,
            self.description,
            self.get_uri(show_password=False),
        )

    def __eq__(self, other: Any) -> bool:
        """Test for equality."""
        if not isinstance(other, ProjectConfig):
            return False
        if other.type != self.type:
            return False
        if other.get_uri(show_password=True) != self.get_uri(show_password=True):
            return False
        if other.description != self.description:
            return False
        if other.project_password != self.project_password:
            return False
        return True

    def load_projectxml(self) -> bool:
        """Collect the connection information from an xml file."""

        file_name = self.filename
        if not os.path.isfile(file_name):
            raise ValueError("El proyecto %r no existe." % file_name)

        tree = ELT.parse(file_name)
        root = tree.getroot()
        version = utils_version.VersionNumber(root.get("Version"), default="1.0")
        self.version = version
        self.description = ""
        for xmldescription in root.findall("name"):
            self.description = xmldescription.text or ""
        profile_pwd = ""
        for profile in root.findall("profile-data"):
            profile_pwd = getattr(profile.find("password"), "text", "")
            if profile_pwd:
                break

        self.password_required = True
        self.checkProfilePasswordForVersion(self.project_password, profile_pwd, version)

        self.fernet = None
        if self.project_password and self.version > VERSION_1_1:
            key_salt = hashlib.sha256(profile_pwd.encode()).digest()
            key = hashlib.pbkdf2_hmac("sha256", self.project_password.encode(), key_salt, 10000)
            key64 = base64.urlsafe_b64encode(key)
            self.fernet = fernet.Fernet(key64)

        from pineboolib.application.database import pnsqldriversmanager

        sql_drivers_manager = pnsqldriversmanager.PNSqlDriversManager()
        self.database = self.retrieveCipherSubElement(root, "database-name")
        for item in root.findall("database-server"):
            self.host = self.retrieveCipherSubElement(item, "host")
            port_text = self.retrieveCipherSubElement(item, "port")
            self.port = int(port_text) if port_text else None
            self.type = self.retrieveCipherSubElement(item, "type")

            # FIXME: Move this to project, or to the connection handler.
            if self.type not in sql_drivers_manager.aliasList():
                LOGGER.warning("Esta versión de pineboo no soporta el driver '%s'" % self.type)

        for credentials in root.findall("database-credentials"):
            self.username = self.retrieveCipherSubElement(credentials, "username")
            self.password = self.retrieveCipherSubElement(credentials, "password")
            if self.password and self.fernet is None:
                self.password = base64.b64decode(self.password).decode()
        self.password_required = False

        return True

    @classmethod
    def encodeProfilePasswordForVersion(
        cls,
        password: str,
        version_: "utils_version.VersionNumber",  # type: ignore [name-defined] # noqa: F821
    ) -> str:
        """
        Hash a password for a profile/projectconfig using the protocol for specified version.
        """
        if password == "":
            return ""
        if version_ < VERSION_1_1:
            return password

        if version_ < VERSION_1_2:
            return hashlib.sha256(password.encode()).hexdigest()
        # Minimum salt size recommended is 8 bytes
        # multiple of 3 bytes as it is shorter in base64
        salt = os.urandom(9)
        hmac = hashlib.pbkdf2_hmac("sha256", password.encode(), salt, 10000)
        dict_passwd = {
            # Algorithm:
            # .. pbkdf2: short algorithm name
            # .. sha256: hash function used
            # .. 4: number of zeroes used on iterations
            "algorithm": "pbkdf2-sha256-4",
            "salt": base64.b64encode(salt).decode(),
            "hash": base64.b64encode(hmac).decode(),
        }
        hashed_password = "%(algorithm)s:%(salt)s:%(hash)s" % dict_passwd

        return hashed_password

    @classmethod
    def checkProfilePasswordForVersion(
        cls,
        user_pwd: str,
        profile_pwd: str,
        version_: "utils_version.VersionNumber",  # type: ignore [name-defined] # noqa: F821
    ) -> None:
        """
        Check a saved password against a user-supplied one.

        user_pwd: User-supplied password in clear text.
        profile_pwd: Raw data saved as password in projectconfig file.
        version: Version number used for checks.

        This function returns None and raises PasswordMismatchError if the password is wrong.
        We can only check if it is good. It's not a good idea to check if two encoded passwords
        are the same, because most secure methods will store different encoded versions every
        time we try to encode again.
        """
        if not profile_pwd:
            return

        if version_ < VERSION_1_1:
            if user_pwd == profile_pwd:
                return
            raise PasswordMismatchError("La contraseña es errónea")

        if version_ < VERSION_1_2:
            user_hash = hashlib.sha256(user_pwd.encode()).hexdigest()
            if profile_pwd == user_hash:
                return
            raise PasswordMismatchError("La contraseña es errónea")

        algo, *algo_extra = profile_pwd.split(":")
        if algo != "pbkdf2-sha256-4":
            raise Exception("Unsupported password algorithm %r" % algo)

        salt64, hash64 = algo_extra

        salt = base64.b64decode(salt64.encode())

        user_hash2 = hashlib.pbkdf2_hmac("sha256", user_pwd.encode(), salt, 10000)
        user_hash64 = base64.b64encode(user_hash2).decode()
        if user_hash64 == hash64:
            return

        raise PasswordMismatchError("La contraseña es errónea")

    def createCipherSubElement(
        self, parent: "ELT.Element", tagname: str, text: str
    ) -> "ELT.Element":
        """Create a XML SubElement ciphered if self.fernet is present."""
        child = ELT.SubElement(parent, tagname)
        if self.fernet is None:
            child.text = text
            return child
        # NOTE: This method returns ciphertext even for empty strings! This is intended.
        # ... this is to avoid anyone knowing if a field is empty or not.
        if len(text) < 64:
            # Right Pad with new line at least up to 64 bytes. Avoid giving out field size.
            text = text.ljust(64, "\n")
        encoded_bytes = self.fernet.encrypt(text.encode())
        encoded_text = base64.b64encode(encoded_bytes).decode()
        child.set("cipher-method", "cryptography.Fernet")
        child.set("cipher-text", encoded_text)
        return child

    def retrieveCipherSubElement(self, parent: "ELT.Element", tagname: str) -> str:
        """Get a XML SubElement ciphered if self.fernet is present."""
        child = parent.find(tagname)
        if child is None:
            raise ValueError("Tag %r not present" % tagname)

        cipher_method = child.get("cipher-method")
        if cipher_method is None:
            return child.text or ""
        if self.fernet is None:
            raise Exception("Tried to load ciphered tag %r with no loaded cipher" % tagname)
        cipher_text = child.get("cipher-text")
        if cipher_method != "cryptography.Fernet":
            raise ValueError("Cipher method %r not supported." % cipher_method)

        if not cipher_text:
            raise ValueError("Missing ciphertext for %r" % tagname)

        cipher_bytes = base64.b64decode(cipher_text.encode())
        text = self.fernet.decrypt(cipher_bytes).decode()
        text = text.rstrip("\n")
        return text

    def save_projectxml(self, overwrite_existing: bool = True) -> None:
        """
        Save the connection.
        """
        profile = ELT.Element("Profile")
        profile.set("Version", str(self.SAVE_VERSION))
        description = self.description
        filename = self.filename
        if not os.path.exists(self.profile_dir):
            os.mkdir(self.profile_dir)

        if not overwrite_existing and os.path.exists(filename):
            raise ProfileAlreadyExistsError

        passw_db = self.password or ""

        profile_user = ELT.SubElement(profile, "profile-data")
        profile_password = ELT.SubElement(profile_user, "password")
        profile_password.text = self.encodeProfilePasswordForVersion(
            self.project_password, self.SAVE_VERSION
        )

        self.fernet = None
        if self.project_password and self.SAVE_VERSION > VERSION_1_1:
            key_salt = hashlib.sha256(profile_password.text.encode()).digest()
            key = hashlib.pbkdf2_hmac("sha256", self.project_password.encode(), key_salt, 10000)
            key64 = base64.urlsafe_b64encode(key)
            self.fernet = fernet.Fernet(key64)
        else:
            # Mask the password if no cipher is used!
            passw_db = base64.b64encode(passw_db.encode()).decode()

        name = ELT.SubElement(profile, "name")
        name.text = description
        dbs = ELT.SubElement(profile, "database-server")
        self.createCipherSubElement(dbs, "type", text=self.type)
        self.createCipherSubElement(dbs, "host", text=self.host or "")
        self.createCipherSubElement(dbs, "port", text=str(self.port) if self.port else "")

        dbc = ELT.SubElement(profile, "database-credentials")
        self.createCipherSubElement(dbc, "username", text=self.username or "")
        self.createCipherSubElement(dbc, "password", text=passw_db)

        self.createCipherSubElement(profile, "database-name", text=self.database)

        utils_base.pretty_print_xml(profile)

        tree = ELT.ElementTree(profile)

        tree.write(filename, xml_declaration=True, encoding="utf-8")
        self.version = self.SAVE_VERSION

    @classmethod
    def translate_connstring(cls, connstring: str) -> Tuple[str, str, str, str, int, str]:
        """
        Translate a DSN connection string into user, pass, etc.

        Accept a "connstring" parameter that has the form user @ host / dbname
        and returns all parameters separately. It takes into account the
        default values ​​and the different abbreviations that exist.
        """
        user = "postgres"
        passwd = ""
        host = "127.0.0.1"
        port = "5432"
        driver_alias = "PostgreSQL (PSYCOPG2)"
        user_pass = None
        host_port = None

        import re

        if "/" not in connstring:
            dbname = connstring
            if not re.match(r"\w+", dbname):
                raise ValueError("base de datos no valida")
            return user, passwd, driver_alias, host, int(port), dbname

        uphpstring = connstring[: connstring.rindex("/")]
        dbname = connstring[connstring.rindex("/") + 1 :]
        up_, hp_ = uphpstring.split("@")
        conn_list = [None, None, up_, hp_]
        _user_pass, _host_port = conn_list[-2], conn_list[-1]

        if _user_pass:
            user_pass = _user_pass.split(":") + ["", "", ""]
            user, passwd, driver_alias = (
                user_pass[0],
                user_pass[1] or passwd,
                user_pass[2] or driver_alias,
            )
            if user_pass[3]:
                raise ValueError("La cadena de usuario debe tener el formato user:pass:driver.")

        if _host_port:
            host_port = _host_port.split(":") + [""]
            host, port = host_port[0], host_port[1] or port
            if host_port[2]:
                raise ValueError("La cadena de host debe ser host:port.")

        if not re.match(r"\w+", user):
            raise ValueError("Usuario no valido")
        if not re.match(r"\w+", dbname):
            raise ValueError("base de datos no valida")
        if not re.match(r"\d+", port):
            raise ValueError("puerto no valido")
        LOGGER.debug(
            "user:%s, passwd:%s, driver_alias:%s, host:%s, port:%s, dbname:%s",
            user,
            "*" * len(passwd),
            driver_alias,
            host,
            port,
            dbname,
        )
        return user, passwd, driver_alias, host, int(port), dbname


class ProfileAlreadyExistsError(Exception):
    """Report that project will not be overwritten."""


class PasswordMismatchError(Exception):
    """Provided password is wrong."""



"""Fltranslator module."""
# -*- coding: utf-8 -*-

import os
from pineboolib.core.utils.utils_base import filedir
from pineboolib.core.utils import logging
from pineboolib.core import settings
from pineboolib import application
from pineboolib.core.utils import utils_base

from PyQt6 import QtCore  # type: ignore[import]


from typing import Dict, Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from PyQt6 import QtWidgets  # noqa: F401 # pragma: no cover

LOGGER = logging.get_logger(__name__)


class PNTranslator(QtCore.QTranslator):
    """FLTranspator class."""

    _multi_lang: bool
    _sys_trans: bool
    _id_module: str
    _lang: str
    _translation_from_qm: bool
    _ts_translation_contexts: Dict[str, Dict[str, str]]

    def __init__(
        self,
        parent: Optional["QtWidgets.QWidget"] = None,
        name: Optional[str] = None,
        multi_lang: bool = False,
        sys_trans: bool = False,
    ) -> None:
        """Inicialize."""
        super().__init__()

        self._prj = parent
        if not name:
            raise Exception("Name is mandatory")
        self._id_module = name[: name.rfind("_")]
        self._lang = name[name.rfind("_") + 1 :]
        self._multi_lang = multi_lang
        self._sys_trans = sys_trans
        self._ts_translation_contexts = {}
        self._translation_from_qm = settings.CONFIG.value(
            "ebcomportamiento/translations_from_qm", False
        )

    def loadTsContent(self, key: str) -> bool:
        """
        Load the contents of an existing translation file into the disk cache into the translator.

        The file must be entered in the disk cache before calling this method, in
        Otherwise, nothing will be done.

        @param key Sha1 key that identifies the file in the disk cache
        @return TRUE if the operation was successful
        """
        if self._id_module == "sys":
            ts_file = filedir("./system_module/translations/%s.%s" % (self._id_module, self._lang))
        else:
            if application.PROJECT.conn_manager is None:
                raise Exception("Project is not connected yet")
            ts_file = filedir(
                "%s/cache/%s/%s/file.ts/%s.%s/%s"
                % (
                    application.PROJECT.tmpdir,
                    application.PROJECT.conn_manager.useConn("default").DBName(),
                    self._id_module,
                    self._id_module,
                    self._lang,
                    key,
                )
            )
        # qmFile = self.AQ_DISKCACHE_DIRPATH + "/" + key + ".qm"

        ret_ = False
        if not self._translation_from_qm:
            ret_ = self.load_ts("%s.ts" % ts_file)
            if not ret_:
                LOGGER.warning("For some reason, i cannot load '%s.ts'", ts_file)
        else:
            qm_file = "%s.qm" % ts_file
            if os.path.exists(qm_file):
                if ts_file in (None, ""):
                    return False

            # else:

            #    trans = pntranslations.PNTranslations()
            #    trans.lrelease("%s.ts" % ts_file, qm_file, not self._multi_lang)

            ret_ = self.load(qm_file)
            if not ret_:
                LOGGER.warning("For some reason, i cannot load '%s'", qm_file)

        return ret_

    def translate(  # type: ignore [override] # noqa: F821
        self, context: str, source_text: str, disambiguation: Optional[str] = None, num: int = -1
    ) -> Optional[str]:
        """Return a translated text."""

        if context.endswith("PlatformTheme"):
            context = "QMessageBox"
        ret_: Optional[str] = None
        if self._translation_from_qm:
            ret_ = super().translate(context, source_text)
            if ret_ == "":
                ret_ = None
        else:
            if context in self._ts_translation_contexts.keys():
                if source_text in self._ts_translation_contexts[context]:
                    ret_ = self._ts_translation_contexts[context][source_text]

        return ret_

    def load_ts(self, file_name: str) -> bool:
        """Load a translation file from a path."""

        try:
            root_ = utils_base.load2xml(file_name)
            for context in root_.findall("context"):
                name_elem = context.find("name")
                if name_elem is None:
                    LOGGER.warning("load_ts: <name> not found, skipping")
                    continue
                context_dict_key = name_elem.text
                if not context_dict_key:
                    continue
                if context_dict_key not in self._ts_translation_contexts.keys():
                    self._ts_translation_contexts[context_dict_key] = {}
                for message in context.findall("message"):
                    translation_elem, source_elem = (
                        message.find("translation"),
                        message.find("source"),
                    )
                    translation_text = translation_elem is not None and translation_elem.text
                    source_text = source_elem is not None and source_elem.text

                    if translation_text and source_text:
                        self._ts_translation_contexts[context_dict_key][
                            source_text
                        ] = translation_text

            return True
        except Exception:
            return False



"""PNConnection_manager module."""
from PyQt6 import QtCore, QtWidgets  # type: ignore[import]

from pineboolib.core.utils import logging, utils_base

from pineboolib.core import garbage_collector, decorators
from pineboolib import application
from pineboolib.interfaces import iconnection
from pineboolib.application.database import pnconnection
from pineboolib.application.database import pnsqlcursor
from pineboolib.application.database import pnsqldriversmanager

from sqlalchemy import exc  # type: ignore [import]
import threading

from typing import Dict, Union, List, Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from pineboolib.fllegacy import flmanager  # pragma: no cover
    from pineboolib.fllegacy import flmanagermodules  # pragma: no cover
    from pineboolib.interfaces import isession  # pragma: no cover

LOGGER = logging.get_logger(__name__)


class PNConnectionManager(QtCore.QObject):
    """PNConnectionManager Class."""

    _drivers_sql_manager: "pnsqldriversmanager.PNSqlDriversManager"
    _manager: Optional["flmanager.FLManager"]
    _manager_modules: Optional["flmanagermodules.FLManagerModules"]
    connections_dict: Dict[str, "pnconnection.PNConnection"] = {}
    limit_connections: int = 0  # Limit of connections to use.
    connections_time_out: int = 0  # Seconds to wait to eliminate the inactive connections.

    REMOVE_CONNECTIONS_AFTER_ATOMIC: bool = False
    SAFE_TIME_SLEEP: float
    safe_mode_level: int

    def __init__(self):
        """Initialize."""

        super().__init__()
        self.connections_dict = {}
        self._manager = None
        self._manager_modules = None
        self.REMOVE_CONNECTIONS_AFTER_ATOMIC = False  # pylint: disable=invalid-name
        self.SAFE_TIME_SLEEP = 0.01  # pylint: disable=invalid-name
        self.safe_mode_level = 0
        self._drivers_sql_manager = pnsqldriversmanager.PNSqlDriversManager()

        LOGGER.debug("Initializing PNConnection Manager:")
        LOGGER.debug(
            "Limit : %s, Time out: %s. (0 disabled)",
            self.limit_connections,
            self.connections_time_out,
        )

    def setMainConn(self, main_conn: "pnconnection.PNConnection") -> bool:
        """Set main connection."""
        if "main_conn" in self.connections_dict:
            if main_conn.conn is not self.connections_dict["main_conn"].conn:
                self.connections_dict["main_conn"].close()
                del self.connections_dict["main_conn"]

        pnsqlcursor.CONNECTION_CURSORS.clear()

        main_conn._name = "main_conn"
        if self._drivers_sql_manager.loadDriver(main_conn._driver_name):
            main_conn.conn = main_conn.conectar(
                main_conn._db_name,
                main_conn._db_host,
                main_conn._db_port,
                main_conn._db_user_name,
                main_conn._db_password,
            )
            if isinstance(main_conn.conn, bool):
                return False

            main_conn._is_open = True

        self.connections_dict["main_conn"] = main_conn
        return True

    def mainConn(self) -> "pnconnection.PNConnection":
        """Return main conn."""

        if "main_conn" in self.connections_dict.keys():
            return self.connections_dict["main_conn"]
        else:
            raise Exception("main_conn is empty!")

    def remove_session(self, session: "isession.PinebooSession") -> bool:
        """Remove session."""

        try:
            session.close()
            obj_ = session
            del session
            garbage_collector.check_delete(obj_, str(obj_))
        except Exception as error:
            LOGGER.warning("Error removing session:%s", error)
            return False

        return True

    def finish(self) -> None:
        """Set the connection as terminated."""

        for key in list(self.connections_dict.keys()):
            if self.connections_dict[key] is None:
                continue

            self.connections_dict[key].close()
            del self.connections_dict[key]

        self.connections_dict.clear()
        del self._manager
        del self._manager_modules
        del self

    def useConn(
        self, name_or_conn: Union[str, "iconnection.IConnection"] = "default", db_name: str = ""
    ) -> "iconnection.IConnection":
        """
        Select another connection which can be not the default one.

        Allow you to select a connection.
        """

        name: str = (
            name_or_conn.connectionName()
            if isinstance(name_or_conn, iconnection.IConnection)
            else name_or_conn
        )

        name_conn_: str = utils_base.session_id(name)
        self.check_alive_connections()

        if name_conn_ in self.connections_dict.keys() and not db_name:
            return self.connections_dict[name_conn_]
        else:
            main_conn = self.mainConn()
            if main_conn is None:
                raise Exception("main_conn is empty!!")

            if name == "main_conn":
                return main_conn
            else:
                if db_name:
                    if not self.removeConn(name):
                        raise Exception("a problem existes deleting older connection")

                new_conn: "pnconnection.PNConnection" = pnconnection.PNConnection(
                    db_name or main_conn._db_name
                )
                new_conn._name = name

                if name.lower() in ["default", "dbaux", "aux"]:  # Las abrimos automáticamene!
                    if self._drivers_sql_manager.loadDriver(new_conn._driver_name):
                        new_conn.conn = new_conn.conectar(
                            new_conn._db_name,
                            new_conn._db_host,
                            new_conn._db_port,
                            new_conn._db_user_name,
                            new_conn._db_password,
                        )
                        new_conn._is_open = True

                self.connections_dict[name_conn_] = new_conn

                return new_conn

    def enumerate(self) -> Dict[str, "pnconnection.PNConnection"]:
        """Return dict with own database connections."""

        dict_ = {}
        id_thread = threading.current_thread().ident
        for key, value in self.connections_dict.items():
            connection_data = key.split("|")
            if connection_data[0] == str(id_thread):
                dict_[connection_data[1]] = value

        return dict_

    def removeConn(self, name="default") -> bool:
        """Delete a connection specified by name."""

        name_conn_: str = utils_base.session_id(name) if name.find("|") == -1 else name

        result = True

        if name_conn_ in self.connections_dict.keys():
            self.connections_dict[name_conn_]._is_open = False

            if self.connections_dict[name_conn_].conn not in [None, self.mainConn().conn]:
                try:
                    if application.SHOW_CONNECTION_EVENTS:
                        LOGGER.info("Closing connection %s", name_conn_)
                    self.connections_dict[name_conn_].close()
                    self.connections_dict[
                        name_conn_
                    ].driver().db_ = None  # type: ignore [assignment]
                    self.connections_dict[name_conn_]._driver = None

                    obj_ = self.connections_dict[name_conn_]
                    garbage_collector.check_delete(obj_, name_conn_)

                except Exception:
                    LOGGER.warning("Connection %s failed when close", name_conn_.split("|")[1])
                    result = False

            self.connections_dict[name_conn_] = None  # type: ignore [assignment] # noqa: F821
            del self.connections_dict[name_conn_]

        return result

    def manager(self) -> "flmanager.FLManager":
        """
        Flmanager instance that manages the connection.

        Flmanager manages metadata of fields, tables, queries, etc .. to then be managed this data by the controls of the application.
        """

        if self._manager is None:
            from pineboolib.fllegacy import flmanager

            self._manager = flmanager.FLManager(self.mainConn())

        return self._manager

    def managerModules(self) -> "flmanagermodules.FLManagerModules":
        """
        Instance of the FLManagerModules class.

        Contains functions to control the state, health, etc ... of the database tables.
        """

        if self._manager_modules is None:
            from pineboolib.fllegacy.flmanagermodules import FLManagerModules

            self._manager_modules = FLManagerModules(self.mainConn())

        return self._manager_modules

    def db(self) -> "iconnection.IConnection":
        """Return the connection itself."""

        return self.useConn("default")

    def dbAux(self) -> "iconnection.IConnection":
        """
        Return the auxiliary connection to the database.

        This connection is useful for out of transaction operations.
        """
        return self.useConn("dbAux")

    def default(self) -> "iconnection.IConnection":
        """
        Return the default connection to the database.
        """
        return self.useConn("default")

    def test_session(
        self, conn_or_session: Union["iconnection.IConnection", "isession.PinebooSession"]
    ) -> bool:
        """Test a specific connection."""

        result = True
        session: "isession.PinebooSession" = (
            conn_or_session.session(False)  # type: ignore [assignment]
            if isinstance(conn_or_session, pnconnection.PNConnection)
            else conn_or_session
        )
        try:
            session.execute("SELECT 1").fetchone()  # type: ignore [arg-type]
            result = hasattr(session, "commit")
        except Exception as error:
            session_name = session._conn_name  # type: ignore [attr-defined] # noqa: F821
            LOGGER.info("Connection %s is bad. error: %s", session_name, str(error))
            result = False

        return result

    def check_connections(self) -> bool:
        """Check connections."""
        self.default()
        self.dbAux()
        for conn_name in list(self.enumerate().keys()):  # Comprobamos conexiones una a una
            conn_identifier = utils_base.session_id(conn_name)
            if conn_identifier in self.connections_dict.keys():
                conn_ = self.connections_dict[conn_identifier]
                LOGGER.debug("Checking connection %s", conn_identifier)
                valid = True
                if not conn_.isOpen():
                    LOGGER.debug("Connection %s is closed.", conn_identifier)
                    valid = False
                else:
                    if not self.test_session(conn_):
                        valid = False

                if not valid:
                    if not self.removeConn(conn_identifier):
                        LOGGER.debug("Connection %s removing failed!", conn_identifier)
                        return False

        return True

    def reinit_user_connections(self) -> None:
        """Reinit users connection."""

        for conn_name in self.enumerate().keys():
            if self.removeConn(conn_name):
                self.useConn(conn_name)

    def check_alive_connections(self):
        """Check alive connections."""

        alived_threads: List[str] = [str(thread.ident) for thread in threading.enumerate()]

        for conn_ident in list(self.connections_dict.keys()):
            if conn_ident.find("|") > -1:
                thread_id = conn_ident.split("|")[0]
                if (
                    conn_ident in self.connections_dict.keys()
                ):  # recompruebo porque puede no existir ya.
                    if (
                        thread_id not in alived_threads  # si no es un hilo existente
                        or not self.connections_dict[conn_ident]._is_open  # si esta cerrada
                        and self.connections_dict[conn_ident].conn
                        is not None  # si no esta incializada
                    ):
                        self.removeConn(conn_ident)

    def set_max_connections_limit(self, limit: int) -> None:
        """Set maximum connections limit."""
        LOGGER.info("New max connections limit %s.", limit)
        self.limit_connections = limit  # noqa: F841

    def set_max_idle_connections(self, limit: int) -> None:
        """Set maximum connections time idle."""
        LOGGER.info("New max connections idle time %s.", limit)
        self.connections_time_out = limit  # noqa: F841

    def active_pncursors(self, only_name: bool = False, all_sessions: bool = False) -> List[str]:
        """Return a user cursor opened list."""

        QtWidgets.QApplication.processEvents()

        identifier = self.session_id()

        result = []

        for key in pnsqlcursor.CONNECTION_CURSORS.keys():
            if key == identifier or all_sessions:
                for cursor_name in pnsqlcursor.CONNECTION_CURSORS[key]:
                    result.append(cursor_name.split("@")[0] if only_name else cursor_name)

        return result

    def session_id(self) -> str:
        """Return session identifier."""

        return application.PROJECT.session_id()

    def is_valid_session(
        self,
        session_or_id: Optional[Union[str, "isession.PinebooSession"]],
        raise_error: bool = True,
    ) -> bool:
        """Return if a session id is valid."""
        is_valid = False
        if application.AUTO_RELOAD_BAD_CONNECTIONS:
            raise_error = False

        session = None

        if session_or_id is not None:
            session = (
                self.useConn(session_or_id).session()
                if isinstance(session_or_id, str)
                else session_or_id
            )

        if session is not None:
            try:
                try:
                    if not session.connection().closed:
                        is_valid = True
                except exc.InvalidRequestError:
                    if session.transaction is None:
                        is_valid = True
                except AttributeError:
                    if session.transaction is None:
                        is_valid = True

            except Exception as error:
                if raise_error:
                    LOGGER.warning(
                        "AttributeError:: Quite possibly, you are trying to use a session in which"
                        " a previous error has occurred and has not"
                        " been recovered with a rollback. Current session is discarded."
                    )
                    raise error

            if application.AUTO_RELOAD_BAD_CONNECTIONS:
                need_reload = False
                if is_valid:
                    if not self.test_session(session):
                        need_reload = True
                        is_valid = False

                if need_reload:
                    LOGGER.warning(
                        "AUTO RELOAD: bad connection detected. Reloading users connections"
                    )
                    if session.transaction is not None:
                        LOGGER.warning(
                            "AUTO RELOAD: bad session %s is currently in transacction. Aborted",
                            session.transaction,
                        )

                    self.reinit_user_connections()

        return is_valid

    def pool_status(self, conn_name: str = "main_conn") -> str:
        """Return pool status used for the conn_name."""

        return self.useConn(conn_name).driver()._engine.pool.status()

    @decorators.deprecated
    def set_safe_mode(self, level: int) -> None:
        """
        Set safe mode level.

        > 0 ) Engine events activated.
        1) Increase time before SERIALIZED calls.
        2) Increase time after SERIALIZED calls.
        3) 1 + 2.
        4) Pool pre pings activated.
        5) 1 + 2 + 4.
        """

        """         self.safe_mode_level = level
        LOGGER.info("CONNECTION MANAGER: Safe mode level set to %s", level)

        if level > 0:
            application.SHOW_CONNECTION_EVENTS = True
            LOGGER.info("CONNECTION MANAGER (%s): Engine events activated.", level)
            LOGGER.info(
                "CONNECTION MANAGER (%s): Pool status when new connection activated.", level
            )

        if level in [1, 3, 5]:
            LOGGER.info(
                "CONNECTION MANAGER (%s): Time before SERIALIZED calls increase %s sec.",
                level,
                self.SAFE_TIME_SLEEP,
            )
        if level in [2, 3, 5]:
            LOGGER.info(
                "CONNECTION MANAGER (%s): Time after SERIALIZED calls increase %s sec.",
                level,
                self.SAFE_TIME_SLEEP,
            )
        if level in [4, 5]:
            LOGGER.info("CONNECTION MANAGER (%s): Pre ping activated.", level) """

    def __getattr__(self, name):
        """Return attributer from main_conn pnconnection."""

        if "main_conn" in self.connections_dict.keys():
            return getattr(self.mainConn(), name, None)

        return None

    database = useConn



"""Qmenu module."""
from PyQt6 import QtCore  # type: ignore[import]

from typing import List, Optional, Any
from pineboolib import application


class QObject(QtCore.QObject):
    """QObject class."""

    _event_filter_function: str
    _allowed_events: List[Any]

    def __ini__(self) -> None:
        """Initialize."""

        super().__init__()

        self._event_filter_function = ""
        self._allowed_events = []

    def eventFilter(
        self, obj: Optional["QtCore.QObject"], event: Optional["QtCore.QEvent"]
    ) -> bool:
        """Return event filter result."""

        result: Any = False
        if self._event_filter_function:
            if event in self._allowed_events:
                result = application.PROJECT.call(self._event_filter_function, [obj, event])
                if not isinstance(result, bool):
                    result = False
        else:
            if event in self._allowed_events:
                result = super().eventFilter(obj, event)  # type: ignore [arg-type]

        return result

    def set_event_filter_function(self, func_name: str) -> None:
        """Set envent finter function name."""

        self._event_filter_function = func_name

    def set_allowed_events(self, allowed_events: List[Any]) -> None:
        """Set allowed events."""

        self._allowed_events = allowed_events

    def get_allowed_events(self) -> List[Any]:
        """Return allowed events lists."""

        return self._allowed_events

    def get_event_filter_function(self) -> str:
        """Return event filter function."""

        return self._event_filter_function

    eventFilterFunction = property(get_event_filter_function, set_event_filter_function)
    allowedEvents = property(get_allowed_events, set_allowed_events)



"""Qframe module."""

# -*- coding: utf-8 -*-
from PyQt6 import QtWidgets  # type: ignore[import]
from typing import Union, TYPE_CHECKING

if TYPE_CHECKING:
    from pineboolib.q3widgets.qgroupbox import QGroupBox  # noqa: F401 # pragma: no cover
    from pineboolib.q3widgets.qwidget import QWidget  # noqa: F401 # pragma: no cover


class QFrame(QtWidgets.QFrame):
    """QFrame class."""

    _line_width: int

    PopupPanel = 7
    MenuBarPanel = 8
    ToolBarPanel = 9
    LineEditPanel = 10
    TabWidgetPanel = 11
    GroupBoxPanel = 12
    MShape = 13

    def __init__(self, parent: Union["QGroupBox", "QWidget"]) -> None:
        """Initialize."""

        super().__init__(parent)
        self.setContentsMargins(0, 0, 0, 0)
        # self.setStyleSheet("QFrame{background-color: transparent}")
        self.setLineWidth(1)
        # self._do_style()

    # def _do_style(self) -> None:
    #    """Set style."""

    #    self.style_str = "QFrame{ background-color: transparent;"
    #    self.style_str += " border-width: %spx;" % self._line_width
    #    self.style_str += " }"
    #    self.setStyleSheet(self.style_str)



"""Flreportengine module."""
from typing import List
from PyQt6 import QtXml, QtCore, QtPrintSupport  # type: ignore

from PyQt6.QtXml import QDomNode as FLDomNodeInterface  # type: ignore # FIXME

from pdf2image import convert_from_path  # type: ignore

from pineboolib.core import decorators, settings
from pineboolib import logging, application
from pineboolib.application.database.pnsqlquery import PNSqlQuery
from typing import Any, Optional, Dict, Union, cast, TYPE_CHECKING

if TYPE_CHECKING:
    from PyQt6 import QtWidgets  # noqa: F401


LOGGER = logging.get_logger(__name__)


class FLReportEnginePrivate(object):
    """FLReportEnginePrivate."""

    rows_: Any
    q_field_mtd_list_: List[Any]
    q_field_list_: List[str]
    q_double_field_list_: List[str]
    q_group_dict_: Dict[int, str]
    q_img_fields_: List[int]
    qry_: Optional[PNSqlQuery]

    def __init__(self, report_engine: "FLReportEngine") -> None:
        """Inicialize."""
        self.qry_ = None
        self.q_field_mtd_list_ = []
        self.q_group_dict_ = {}
        self.report_engine = report_engine
        self.template_ = ""
        self.rows_ = []
        self.q_field_list_ = []
        self.q_double_field_list_ = []
        self.q_img_fields_ = []

    def addRowToReportData(self, level: int) -> None:
        """Add row to report data."""
        if not self.qry_ or not self.qry_.isValid():
            return

        row = self.report_engine.rptXmlData().createElement("Row")
        row.setAttribute("level", level)

        for num, item in enumerate(self.q_field_list_):
            str_val = str(self.qry_.value(item, False))
            if self.q_img_fields_:
                if self.q_img_fields_[-1] == num:
                    self.q_img_fields_.append(self.q_img_fields_.pop())
                    if str_val in ["", "None"]:
                        row.setAttribute(item, str_val)
                        continue

                key = self.qry_.value(num, False)
                str_val = str(key)

            row.setAttribute(item, str_val)

        self.rows_.appendChild(row)

    def groupBy(self, level_max: int, va_: List[Any]) -> None:
        """Add a group by."""
        if not self.qry_ or not self.qry_.isValid():
            return

        # self.q_group_dict_
        lev = 0

        while lev < level_max and str(self.qry_.value(self.q_group_dict_[lev])) == str(va_[lev]):
            lev += 1

        for num in range(lev, level_max):
            self.addRowToReportData(num)
            va_[num] = str(self.qry_.value(self.q_group_dict_[num]))

        self.addRowToReportData(level_max)

    def setQuery(self, qry: Optional[PNSqlQuery]) -> None:
        """Set query to the report."""
        self.qry_ = qry

        if self.qry_:
            self.q_field_list_ = self.qry_.fieldList(True)
            self.q_field_mtd_list_ = self.qry_.fieldMetaDataList()
            self.q_group_dict_ = self.qry_.groupDict()
            self.q_double_field_list_.clear()
            self.q_img_fields_.clear()

            if not self.q_field_mtd_list_:
                return

            size = len(self.q_field_list_) - 1
            while size >= 0:
                item = self.q_field_list_[size]
                key = item[item.find(".") + 1 :].lower()
                for field in self.q_field_mtd_list_:
                    if field.name() == key:
                        if field.type() == "pixmap":
                            self.q_img_fields_.append(size)
                        elif field.type() == "double":
                            self.q_double_field_list_.append(item)
                        break
                size -= 1
        else:
            self.q_field_list_.clear()
            self.q_double_field_list_.clear()
            self.q_img_fields_.clear()
            self.q_field_mtd_list_ = []
            self.q_group_dict_ = {}


class FLReportEngine(QtCore.QObject):
    """FLReportEngine class."""

    report_: Any
    report_template_: Optional[str]
    _rel_dpi: int
    _private: "FLReportEnginePrivate"

    def __init__(self, parent: Optional["QtWidgets.QWidget"] = None) -> None:
        """Inicialize."""

        super().__init__(parent)
        self._private = FLReportEnginePrivate(self)
        self._rel_dpi = int(settings.SETTINGS.value("rptViewer/pixel", 780))
        self.report_ = None
        self.report_template_ = ""
        self.report_data_: Optional[QtXml.QDomDocument] = None

        from pineboolib.application.parsers.parser_kut import kut2fpdf

        self._parser: "kut2fpdf.Kut2FPDF" = kut2fpdf.Kut2FPDF()

    def printReport(
        self,
        name_or_dialog: Union[str, "QtPrintSupport.QPrintDialog"],
        num_copies: int = 1,
        color_mode: int = 1,
    ) -> bool:
        """Print report to a printer."""

        try:
            from PyQt6 import QtGui
            from PIL.ImageQt import ImageQt  # type: ignore [import]

            page_filter: List[int] = []
            if not isinstance(name_or_dialog, str):
                printer = name_or_dialog.printer()
                range_ = name_or_dialog.printRange()
                first = name_or_dialog.fromPage()
                last = name_or_dialog.toPage()
                if range_.value == 2:  # type: ignore [comparison-overlap] # 0 all, 1,selection, 2 range, 3 current page
                    for num in range(first, last + 1):
                        page_filter.append(num)

            else:
                printer = QtPrintSupport.QPrinter()
                printer.setPrinterName(name_or_dialog)
                printer.setColorMode(cast(QtPrintSupport.QPrinter.ColorMode, color_mode))
                if printer.supportsMultipleCopies():
                    printer.setCopyCount(num_copies)
                else:
                    LOGGER.warning("CopyCount not supported by %s", name_or_dialog)

            if printer:
                printer.setResolution(self._rel_dpi)
                printer.setDocName("Document from Viewer")
                printer.setCreator("Pineboo")
            pdf_file = self._parser.get_file_name()

            images = convert_from_path(
                pdf_file, dpi=self._rel_dpi, output_folder=application.PROJECT.tmpdir
            )

            painter = QtGui.QPainter()
            painter.begin(printer)
            first_ = True
            for num, image in enumerate(images):
                page_index = num + 1
                if not page_filter or page_index in page_filter:
                    if not first_:
                        printer.newPage()  # type: ignore [union-attr]
                    first_ = False
                    rect = painter.viewport()
                    image_qt = ImageQt(image)
                    painter.drawImage(rect, image_qt)
            painter.end()

            return True
        except Exception as error:
            LOGGER.warning("Error printing: %s", str(error))

        return False

    def rptXmlData(self) -> Any:
        """Return report Xml Data."""
        return self.report_data_

    def rptXmlTemplate(self) -> Optional[str]:
        """Return report Xml Template."""
        return self.report_template_

    def relDpi(self) -> float:
        """Return dpi size."""
        return self._rel_dpi

    def setReportData(
        self, qry: Optional[Union[FLDomNodeInterface, PNSqlQuery]] = None
    ) -> Optional[bool]:
        """Set data source to report."""

        if isinstance(qry, FLDomNodeInterface):
            return self.setFLReportData(qry)
        if qry is None:
            return None

        self.report_data_ = QtXml.QDomDocument("KugarData")

        self._private.rows_ = (
            self.report_data_.createDocumentFragment()
        )  # FIXME: Don't set the private from the public.
        self._private.setQuery(qry)
        qry.setForwardOnly(True)
        if qry.exec_() and qry.next():
            group = self._private.q_group_dict_
            if not group:
                while True:
                    self._private.addRowToReportData(0)
                    if not qry.next():
                        break
            else:
                values_: List[None] = []
                for item in range(10):
                    values_.append(None)

                ok_ = True
                while ok_:
                    self._private.groupBy(len(group), values_)
                    if not qry.next():
                        ok_ = False

        data = self.report_data_.createElement("KugarData")
        data.appendChild(self._private.rows_)
        self.report_data_.appendChild(data)
        self._private.rows_.clear()

        self.initData()
        return True

    def setFLReportData(self, data: Any) -> bool:
        """Set data to report."""
        self._private.setQuery(None)
        tmp_doc = QtXml.QDomDocument("KugarData")
        tmp_doc.appendChild(data)
        self.report_data_ = tmp_doc
        return True
        # return super(FLReportEngine, self).setReportData(n)

    def setFLReportTemplate(self, template: Any) -> bool:
        """Set template to report."""
        # buscamos el .kut o el .rlab

        self._private.template_ = template

        if not self._private.qry_:
            from pineboolib import application

            if application.PROJECT.conn_manager is None:
                raise Exception("Project is not connected yet")
            mgr = application.PROJECT.conn_manager.managerModules()

        else:
            mgr = self._private.qry_.db().connManager().managerModules()

        self.report_template_ = mgr.contentCached("%s.kut" % template)

        if not self.report_template_:
            LOGGER.error("FLReportEngine::No se ha podido cargar %s.kut", template)
            return False

        return True

    def rptQueryData(self) -> Optional[PNSqlQuery]:
        """Return report query data."""
        return self._private.qry_

    def rptNameTemplate(self) -> str:
        """Return report template name."""
        return self._private.template_

    @decorators.beta_implementation
    def setReportTemplate(self, template: Any):
        """Set template to report."""

        if isinstance(template, FLDomNodeInterface):
            return self.setFLReportData(template)

        return self.setFLReportData(template)

    def reportData(self) -> Any:
        """Return report data."""
        return self.report_data_ if self.report_data_ else QtXml.QDomDocument()

    def reportTemplate(self) -> Any:
        """Return report template."""
        return self.report_template_ if self.report_template_ else QtXml.QDomDocument()

    @decorators.not_implemented_warn
    def csvData(self) -> str:
        """Return csv data."""
        # FIXME: Should return the report converted to CSV
        return ""

    @decorators.not_implemented_warn
    def exportToOds(self):
        """Return report exported to odf."""

        return

    def renderReport(
        self, init_row: int = 0, init_col: int = 0, flags: List[int] = [], pages: Any = None
    ) -> "QtCore.QObject":
        """Render report."""
        if (
            self.report_data_
            and self.report_template_
            and self.report_template_.find("KugarTemplate") > -1
        ):
            data = self.report_data_.toString(1)
            self.report_ = self._parser.parse(
                self._private.template_, self.report_template_, data, self.report_, flags
            )

        return QtCore.QObject()  # return self.pages!

        # # print(self.report_data_.toString(1))
        # """
        # fr = MReportEngine.RenderReportFlags.FillRecords.value
        #
        # pgs = FLReportPages()
        # if pages:
        #     pgs.setPageCollection(pages)
        #
        # pgc = super(FLReportEngine, self).renderReport(
        #     initRow,
        #     initCol,
        #     pgs,
        #     fr if fRec else 0
        # )
        #
        # pgs.setPageCollection(pgc)
        # if not fRec or not self._private.qry_ or not self._private.q_field_mtd_list_ or not self._private.q_double_field_list_:
        #     return pgs
        #
        # nl = QtXml.QDomNodeList(self.report_data_.elementsByTagName("Row"))
        # for i in range(nl.count()):
        #     itm = nl.item(i)
        #     if itm.isNull():
        #         continue
        #     nm = itm.attributes()
        #
        #     for it in self._private.q_double_field_list_:
        #         ita = nm.namedItem(it)
        #         if ita.isNull():
        #             continue
        #         sVal = ita.nodeValue()
        #         if not sVal or sVal == "" or sVal.upper() == "NAN":
        #             continue
        #         dVal = float(sVal)
        #         if not dVal:
        #             dVal = 0
        #         decimals = self._private.q_field_mtd_list_.find(
        #             it.section('.', 1, 1).lower()).partDecimal()
        #         ita.setNodeValue(FLUtil.formatoMiles(round(dVal, decimals)))
        # return pgs
        # """

    def initData(self) -> None:
        """Inialize data."""
        if not self.report_data_:
            raise Exception("RD is missing. Initialize properly before calling initData")
        child = self.report_data_.firstChild()
        while not child.isNull():
            if child.nodeName() == "KugarData":
                self.records_ = child.childNodes()
                attr = child.attributes()
                tempattr = attr.namedItem("Template")
                tempname = tempattr.nodeValue() or None
                if tempname is not None:
                    # FIXME: We need to add a signal:
                    # self.preferedTemplate.emit(tempname)
                    break
            child = child.nextSibling()

    def number_pages(self) -> int:
        """Return page numbers."""
        return self._parser.number_pages() if self._parser else 0



"""Test_flreportengine module."""

import unittest
from pineboolib.loader.main import init_testing, finish_testing
from pineboolib.qsa import qsa


class TestFLReportEngine(unittest.TestCase):
    """TestSysType Class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_report_data(self) -> None:
        """Test eneboopkgs load."""

        qry = qsa.FLSqlQuery()
        qry.setTablesList("flfiles")
        qry.setSelect("count(nombre)")
        qry.setFrom("flfiles")
        qry.setWhere("1=1")
        self.assertTrue(qry.exec_())
        self.assertTrue(qry.first())
        res = qsa.sys.toXmlReportData(qry)
        self.assertTrue(res.toString(2).find("</KugarData>") > -1)

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()



"""Test_pnsqlcursor module."""

import unittest
from pineboolib.loader.main import init_testing, finish_testing
from pineboolib.core.utils import logging
from pineboolib.application.database import pnsqlcursor
from pineboolib.application.metadata import pnrelationmetadata
from pineboolib import application
from pineboolib.application.database.tests import fixture_path
from pineboolib.core.utils import utils_base

LOGGER = logging.get_logger("test")


class TestInsertData(unittest.TestCase):
    """TestInsertData Class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        utils_base.FORCE_DESKTOP = True
        init_testing()

    def test_basic_1(self) -> None:
        """Insert data into a database."""

        cursor = pnsqlcursor.PNSqlCursor("flareas")
        cursor.setModeAccess(cursor.Insert)
        cursor.refreshBuffer()
        cursor.setValueBuffer("bloqueo", True)
        cursor.setValueBuffer("idarea", "T")
        cursor.setValueBuffer("descripcion", "Área de prueba T")
        self.assertTrue(cursor.commitBuffer())
        self.assertEqual(cursor.valueBuffer("idarea"), "T")
        mode_access = cursor.modeAccess()
        self.assertEqual(mode_access, cursor.Edit)
        cursor.setModeAccess(cursor.Insert)
        cursor.refreshBuffer()
        cursor.setValueBuffer("bloqueo", True)
        cursor.setValueBuffer("idarea", "T2")
        cursor.setValueBuffer("descripcion", "Área de prueba T2")
        self.assertTrue(cursor.commitBuffer())
        self.assertEqual(cursor.valueBuffer("idarea"), "T2")
        cursor.select("idarea = 'T2'")
        self.assertTrue(cursor.first())
        cursor.setModeAccess(cursor.Del)
        cursor.refreshBuffer()
        cursor.commitBuffer()
        self.assertEqual(cursor.size(), 0)
        cursor.select()
        self.assertEqual(cursor.size(), 1)
        mode_access = cursor.modeAccess()
        self.assertEqual(mode_access, cursor.Browse)
        self.assertTrue(cursor.first())
        cursor.setModeAccess(cursor.Edit)
        cursor.setValueBuffer("descripcion", " ")
        self.assertTrue(cursor.commitBuffer())

    def test_basic_2(self) -> None:
        """SetAtomitValuebuffer."""

        cursor = pnsqlcursor.PNSqlCursor("flareas")
        cursor.select()
        self.assertTrue(cursor.first())
        cursor.setAtomicValueBuffer("descripcion", "sys.iface.get_description")
        self.assertEqual(cursor.valueBuffer("descripcion"), "Área de prueba T.")

    def test_basic_3(self) -> None:
        """Test None values."""
        from pineboolib.qsa import qsa

        cursor = pnsqlcursor.PNSqlCursor("fltest")
        cursor.setModeAccess(cursor.Insert)
        cursor.refreshBuffer()
        self.assertTrue(cursor.valueBuffer("empty_relation") == "")
        self.assertTrue(cursor.buffer().value("empty_relation", True) is None)
        self.assertTrue(cursor.commitBuffer())
        obj = qsa.orm.fltest.get(cursor.valueBuffer("id"))
        self.assertTrue(obj)
        self.assertEqual(obj.empty_relation, None)
        self.assertEqual(obj.uint_field, None)
        self.assertEqual(obj.bool_field, None)
        self.assertEqual(obj.double_field, 0)
        result = qsa.FLUtil.sqlSelect(
            "fltest", "empty_relation", "id = %s" % cursor.valueBuffer("id")
        )
        self.assertEqual(result, "")

    def test_basic_4(self) -> None:
        """Test cursor with external table_name"""

        cursor = pnsqlcursor.PNSqlCursor("fltest7")
        self.assertTrue(cursor.metadata() is not None)
        self.assertEqual(cursor.metadata().name(), "fltest")

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()


class TestEditData(unittest.TestCase):
    """TestEditData Class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_basic(self) -> None:
        """Edit data from a database."""
        cursor = pnsqlcursor.PNSqlCursor("flareas")
        cursor.setModeAccess(cursor.Insert)
        cursor.refreshBuffer()
        cursor.setValueBuffer("bloqueo", True)
        cursor.setValueBuffer("idarea", "T")
        cursor.setValueBuffer("descripcion", "Área de prueba T")
        self.assertTrue(cursor.commitBuffer())

        cursor.select("idarea ='T'")
        first_result = cursor.first()
        self.assertEqual(first_result, True)
        cursor.setModeAccess(cursor.Edit)
        cursor.refreshBuffer()

        value_idarea = cursor.valueBuffer("idarea")
        self.assertEqual(value_idarea, "T")

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()


class TestDeleteData(unittest.TestCase):
    """TestDeletedata Class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_basic_1(self) -> None:
        """Delete data from a database."""
        cursor = pnsqlcursor.PNSqlCursor("flareas")
        cursor.setModeAccess(cursor.Insert)
        cursor.refreshBuffer()
        cursor.setValueBuffer("bloqueo", True)
        cursor.setValueBuffer("idarea", "T")
        cursor.setValueBuffer("descripcion", "Área de prueba T")
        self.assertTrue(cursor.commitBuffer())

        cursor.select("idarea ='T'")
        first_result = cursor.first()
        self.assertEqual(first_result, True)
        size_1 = cursor.size()
        self.assertEqual(size_1, 1, "tiene que devolver 1 y ha devuelto %s" % size_1)
        cursor.setForwardOnly(True)
        cursor.setModeAccess(cursor.Del)
        cursor.refreshBuffer()

        value_idarea = cursor.valueBuffer("idarea")
        self.assertEqual(value_idarea, "T")
        cursor.commitBuffer()
        cursor.refresh()
        size_2 = cursor.size()
        self.assertEqual(size_2, 1)
        cursor.setForwardOnly(False)
        cursor.refresh()
        size_3 = cursor.size()
        self.assertEqual(size_3, 0, "Tiene que devolver 0 y ha devuelto %s" % size_3)

    def test_basic_2(self) -> None:
        """Delete data from a database."""

        cursor = pnsqlcursor.PNSqlCursor("flareas")
        cursor.setModeAccess(cursor.Insert)
        cursor.refreshBuffer()
        cursor.setValueBuffer("bloqueo", True)
        cursor.setValueBuffer("idarea", "T1")
        cursor.setValueBuffer("descripcion", "Área de prueba T1")
        self.assertTrue(cursor.commitBuffer())
        cursor.setModeAccess(cursor.Insert)
        cursor.refreshBuffer()
        cursor.setValueBuffer("bloqueo", True)
        cursor.setValueBuffer("idarea", "T2")
        cursor.setValueBuffer("descripcion", "Área de prueba T2")
        self.assertTrue(cursor.commitBuffer())
        cursor.setModeAccess(cursor.Insert)
        cursor.refreshBuffer()
        cursor.setValueBuffer("bloqueo", True)
        cursor.setValueBuffer("idarea", "T3")
        cursor.setValueBuffer("descripcion", "Área de prueba T3")
        self.assertTrue(cursor.commitBuffer())
        cursor.setModeAccess(cursor.Insert)
        cursor.refreshBuffer()
        cursor.setValueBuffer("bloqueo", True)
        cursor.setValueBuffer("idarea", "T4")
        cursor.setValueBuffer("descripcion", "Área de prueba T4")
        self.assertTrue(cursor.commitBuffer())
        cursor.select()
        self.assertEqual(cursor.size(), 4)
        pass_ = 0
        borrados = []
        cursor.select()
        cursor.setForwardOnly(True)
        while cursor.next():
            pass_ += 1
            cursor.setModeAccess(cursor.Del)
            borrados.append(cursor.valueBuffer(cursor.primaryKey()))
            cursor.refreshBuffer()
            self.assertTrue(cursor.commitBuffer(False))

        self.assertEqual(
            pass_, 4, "Solo ha borrado %s registros! %s" % (pass_, ", ".join(borrados))
        )

    def test_basic_3(self) -> None:
        """Delete data from a database."""

        # añadir relación fake
        relation_meta = pnrelationmetadata.PNRelationMetaData(
            "fake", "idarea", pnrelationmetadata.PNRelationMetaData.RELATION_1M
        )

        cursor = pnsqlcursor.PNSqlCursor("flareas")

        field = cursor.metadata().field("idarea")
        if field is not None:
            field.addRelationMD(relation_meta)
        else:
            self.assertTrue(False, "No se encuentra el campo idarea (1)")

        # comprobar si el cambio es persistente

        metadata_cursor = cursor.metadata()
        meta_field = metadata_cursor.field("idarea")
        if meta_field is None:
            self.assertTrue(False, "No se encuentra el campo idarea (2)")

        relations = meta_field.relationList() if meta_field else []
        found = False
        for relation in relations:
            if relation.foreignTable() == "fake":
                found = True
                break

        self.assertTrue(found, "No se encuentra la relacion FAKE")

        cursor.setModeAccess(cursor.Insert)
        cursor.refreshBuffer()
        cursor.setValueBuffer("bloqueo", True)
        cursor.setValueBuffer("idarea", "T1")
        cursor.setValueBuffer("descripcion", "Área de prueba T1")
        self.assertTrue(cursor.commitBuffer())

        cursor.select("idarea = 'T1'")
        self.assertTrue(cursor.first())
        cursor.setModeAccess(cursor.Del)
        cursor.refreshBuffer()
        self.assertTrue(cursor.commitBuffer())

    def test_basic_4(self) -> None:
        """Delete with relation errors."""
        cursor = pnsqlcursor.PNSqlCursor("flareas")
        cursor.setModeAccess(cursor.Insert)
        cursor.refreshBuffer()
        cursor.setValueBuffer("bloqueo", True)
        cursor.setValueBuffer("idarea", "T")
        cursor.setValueBuffer("descripcion", "Área de prueba T")
        self.assertTrue(cursor.commitBuffer())

        cursor.select("idarea ='T'")
        first_result = cursor.first()
        self.assertEqual(first_result, True)
        size_1 = cursor.size()
        self.assertEqual(size_1, 1, "tiene que devolver 1 y ha devuelto %s" % size_1)
        cursor.setForwardOnly(True)
        cursor.setModeAccess(cursor.Del)
        cursor.refreshBuffer()

        value_idarea = cursor.valueBuffer("idarea")
        self.assertEqual(value_idarea, "T")
        cursor.commitBuffer()
        cursor.refresh()
        size_2 = cursor.size()
        self.assertEqual(size_2, 1)
        cursor.setForwardOnly(False)
        cursor.refresh()
        size_3 = cursor.size()
        self.assertEqual(size_3, 0, "Tiene que devolver 0 y ha devuelto %s" % size_3)

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()


class TestMove(unittest.TestCase):
    """Test Move class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_basic(self) -> None:
        from pineboolib.application.database import pnsqlcursor

        cursor = pnsqlcursor.PNSqlCursor("flareas")
        cursor.setModeAccess(cursor.Insert)
        cursor.refreshBuffer()
        cursor.setValueBuffer("bloqueo", False)
        cursor.setValueBuffer("idarea", "A")
        cursor.setValueBuffer("descripcion", "Área de prueba A")
        self.assertTrue(cursor.commitBuffer())
        cursor.setModeAccess(cursor.Insert)
        cursor.refreshBuffer()
        cursor.setValueBuffer("bloqueo", False)
        cursor.setValueBuffer("idarea", "B")
        cursor.setValueBuffer("descripcion", "Área de prueba B")
        self.assertTrue(cursor.commitBuffer())
        cursor.setModeAccess(cursor.Insert)
        cursor.refreshBuffer()
        cursor.setValueBuffer("bloqueo", False)
        cursor.setValueBuffer("idarea", "C")
        cursor.setValueBuffer("descripcion", "Área de prueba C")
        self.assertTrue(cursor.commitBuffer())
        cursor.setModeAccess(cursor.Insert)
        cursor.refreshBuffer()
        cursor.setValueBuffer("bloqueo", False)
        cursor.setValueBuffer("idarea", "D")
        cursor.setValueBuffer("descripcion", "Área de prueba D")
        self.assertTrue(cursor.commitBuffer())
        cursor.setModeAccess(cursor.Insert)
        cursor.refreshBuffer()
        cursor.setValueBuffer("bloqueo", False)
        cursor.setValueBuffer("idarea", "E")
        cursor.setValueBuffer("descripcion", "Área de prueba E")
        self.assertTrue(cursor.commitBuffer())
        cursor.select()
        cursor.first()
        self.assertEqual(cursor.valueBuffer("idarea"), "A")
        res_1 = cursor.prev()
        self.assertEqual(res_1, False)
        self.assertEqual(cursor.valueBuffer("idarea"), "A")
        res_2 = cursor.last()
        self.assertEqual(res_2, True)
        self.assertEqual(cursor.valueBuffer("idarea"), "E")
        res_3 = cursor.next()
        self.assertEqual(res_3, False)
        self.assertEqual(cursor.valueBuffer("idarea"), "E")
        res_4 = cursor.prev()
        self.assertEqual(res_4, True)
        self.assertEqual(cursor.valueBuffer("idarea"), "D")
        cursor.prev()
        self.assertEqual(cursor.valueBuffer("idarea"), "C")
        res_5 = cursor.first()
        self.assertEqual(res_5, True)
        self.assertEqual(cursor.valueBuffer("idarea"), "A")
        cursor.next()
        self.assertEqual(cursor.valueBuffer("idarea"), "B")
        res_6 = cursor.next()
        self.assertEqual(res_6, True)
        self.assertEqual(cursor.valueBuffer("idarea"), "C")
        self.assertEqual(cursor.size(), 5)

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()


class TestBuffer(unittest.TestCase):
    """Test buffer class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_basic(self) -> None:
        """Test buffers data."""
        from pineboolib.application.database import pnsqlcursor

        cursor = pnsqlcursor.PNSqlCursor("flareas")
        cursor.setModeAccess(cursor.Insert)
        cursor.refreshBuffer()
        cursor.setValueBuffer("bloqueo", True)
        cursor.setValueBuffer("idarea", "T")
        cursor.setValueBuffer("descripcion", "Área de prueba T")
        self.assertTrue(cursor.commitBuffer())
        cursor.setModeAccess(cursor.Insert)
        cursor.refreshBuffer()
        cursor.setValueBuffer("bloqueo", True)
        cursor.setValueBuffer("idarea", "V")
        cursor.setValueBuffer("descripcion", "Área de prueba V")
        self.assertTrue(cursor.commitBuffer())
        cursor.select()
        self.assertEqual(cursor.size(), 2)
        cursor.select("1=1 ORDER BY idarea ASC")
        cursor.setModeAccess(cursor.Edit)
        self.assertEqual(cursor.size(), 2)
        self.assertTrue(cursor.first())
        cursor.refreshBuffer()
        buffer_copy = cursor.bufferCopy()
        buffer = cursor.buffer()
        self.assertNotEqual(buffer, None)
        self.assertEqual(cursor.valueBuffer("idarea"), "T")
        self.assertEqual(cursor.valueBufferCopy("descripcion"), "Área de prueba T")

        self.assertEqual(buffer.value("idarea"), "T")
        self.assertEqual(buffer_copy.value("idarea"), "T")
        cursor.next()
        buffer = cursor.buffer()
        self.assertEqual(buffer.value("idarea"), "V")
        self.assertEqual(buffer_copy.value("idarea"), "T")

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()


class TestValues(unittest.TestCase):
    """Test Values class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_basic(self) -> None:
        """Test values."""
        from pineboolib.application.database import pnsqlcursor
        from pineboolib.application.qsatypes import date

        cursor = pnsqlcursor.PNSqlCursor("flupdates")
        cursor.setModeAccess(cursor.Insert)
        cursor.refreshBuffer()
        date_ = date.Date()
        cursor.setValueBuffer("fecha", date_)
        cursor.setValueBuffer("hora", "00:00:01")
        cursor.setValueBuffer("nombre", "nombre de prueba")
        cursor.setValueBuffer("modulesdef", "module_1\nmodule_2\nmodule_3")
        cursor.setValueBuffer("filesdef", "file_1\nfile_2\nfile_3")
        cursor.setValueBuffer("shaglobal", "1234567890")
        cursor.setValueBuffer("auxtxt", "aux_1\naux_2\naux_3")
        self.assertEqual(cursor.commitBuffer(), True)
        self.assertEqual(str(cursor.valueBuffer("fecha"))[0:10], str(date_)[0:10])
        self.assertEqual(cursor.valueBuffer("hora"), "00:00:01")
        self.assertEqual(cursor.valueBuffer("nombre"), "nombre de prueba")

        cursor_2 = pnsqlcursor.PNSqlCursor("fltest3")
        cursor_2.setModeAccess(cursor_2.Insert)
        cursor_2.refreshBuffer()
        cursor_2.setValueBuffer("string_field", "Campo de prueba test_pnsqlcursor_test_basic")
        self.assertTrue(cursor_2.commitBuffer())
        cursor_2.select()
        self.assertTrue(cursor_2.first())
        self.assertTrue(cursor_2.valueBuffer("counter_field"), "000001")
        cursor_2.setModeAccess(cursor_2.Del)
        cursor_2.refreshBuffer()
        self.assertTrue(cursor_2.commitBuffer())

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()


class TestEmits(unittest.TestCase):
    """Test Emits class."""

    _transaction_begin: bool
    _transaction_end: bool
    _transaction_roll_back: bool

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_full(self) -> None:
        """Test full."""
        self._transaction_begin = False
        self._transaction_end = False
        self._transaction_roll_back = False

        from pineboolib.application.database import pnsignals
        from pineboolib.application.database import pnsqlcursor

        signals = pnsignals.PNSignals()
        signals.notify_begin_transaction_ = True
        signals.notify_end_transaction_ = True
        signals.notify_roll_back_transaction_ = True

        cursor = pnsqlcursor.PNSqlCursor("test")

        cursor.transactionBegin.connect(self.mark_transaction_begin)
        cursor.transactionEnd.connect(self.mark_transaction_end)
        cursor.transactionRollback.connect(self.mark_transaction_roll_back)

        signals.emitTransactionBegin(cursor)
        signals.emitTransactionEnd(cursor)
        signals.emitTransactionRollback(cursor)

        self.assertTrue(self._transaction_begin)
        self.assertTrue(self._transaction_end)
        self.assertTrue(self._transaction_roll_back)
        self.assertTrue(cursor.rollback())

    def mark_transaction_begin(self) -> None:
        """Mark transaction begin."""

        self._transaction_begin = True

    def mark_transaction_end(self) -> None:
        """Mark transaction end."""

        self._transaction_end = True

    def mark_transaction_roll_back(self) -> None:
        """Mark transaction roll back."""
        self._transaction_roll_back = True

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()


class TestGeneral(unittest.TestCase):
    """Test General class."""

    @classmethod
    def setUp(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        utils_base.FORCE_DESKTOP = True
        init_testing()

    def test_basic_1(self) -> None:
        """Basic tests 1."""
        from pineboolib import application
        from pineboolib.application.database import pnsqlcursor

        cursor = pnsqlcursor.PNSqlCursor("fltest")
        cursor2 = pnsqlcursor.PNSqlCursor(
            "flareas", True, application.PROJECT.conn_manager.mainConn()
        )
        self.assertEqual(cursor.table(), "fltest")
        action = cursor.action()
        action2 = cursor2.action()
        self.assertTrue(action is not None)
        self.assertTrue(action2 is not None)

        self.assertEqual(cursor.actionName(), "fltest")
        if action is not None:
            self.assertTrue(cursor.setAction(action))
        if action2 is not None:
            self.assertTrue(cursor.setAction(action2))

        cursor3 = pnsqlcursor.PNSqlCursor("fltest")
        cursor3.select()
        self.assertEqual(
            cursor3.size(), 0, "el tamaño es %s cuando tendría que ser 0" % cursor3.size()
        )
        cursor3.setMainFilter("id > 1")
        cursor3.select()
        cursor3.refresh()
        self.assertEqual(cursor3.mainFilter(), "id > 1")
        cursor3.refreshBuffer()
        self.assertEqual(cursor3.baseFilter(), "id > 1")

        # self.assertFalse(cursor3.meta_model())
        # self.assertFalse(cursor3.inTransaction())
        # self.assertTrue(cursor3.commit())

        # cursor3.refreshBuffer()

        cursor4 = pnsqlcursor.PNSqlCursor("flareas", "default")

        cursor = pnsqlcursor.PNSqlCursor("flareas")
        cursor.setModeAccess(cursor.Insert)
        cursor.refreshBuffer()
        cursor.setValueBuffer("bloqueo", True)
        cursor.setValueBuffer("idarea", "T")
        cursor.setValueBuffer("descripcion", "Área de prueba T")
        self.assertTrue(cursor.commitBuffer())

        cursor4.select()
        cursor4.first()
        cursor4.setModeAccess(cursor4.Edit)
        cursor4.refreshBuffer()
        self.assertFalse(cursor4.isNull("idarea"))
        self.assertFalse(cursor4.isCopyNull("idarea"))

    def test_basic_2(self) -> None:
        """Basic tests 2."""
        from pineboolib.application.database import pnsqlcursor

        cursor = pnsqlcursor.PNSqlCursor("flareas")
        cursor.setModeAccess(cursor.Insert)
        cursor.refreshBuffer()
        cursor.setValueBuffer("bloqueo", True)
        cursor.setValueBuffer("idarea", "X")
        cursor.setValueBuffer("descripcion", "Área de prueba X")
        self.assertTrue(cursor.commitBuffer())

        cursor = pnsqlcursor.PNSqlCursor("flareas", "default")
        cursor.select()
        cursor.first()
        cursor.setModeAccess(cursor.Edit)

        self.assertFalse(cursor.isModifiedBuffer())
        cursor.setValueBuffer("descripcion", "Descripcion de prueba para a")
        self.assertTrue(cursor.isModifiedBuffer())

        cursor2 = pnsqlcursor.PNSqlCursor("flmodules", "default")
        cursor2.setAskForCancelChanges(False)
        cursor2.setActivatedCheckIntegrity(False)
        cursor2.setActivatedCommitActions(False)

        self.assertFalse(cursor2.activatedCommitActions())
        cursor2.setModeAccess(cursor2.Insert)
        cursor2.refreshBuffer()
        cursor2.setValueBuffer("idmodulo", "Z")
        cursor2.setValueBuffer("idarea", "X")
        cursor2.setValueBuffer("descripcion", "Esta es la descripción")
        cursor2.setValueBuffer("version", "0.0")
        self.assertTrue(cursor2.commitBuffer())

        cursor2.select('idmodulo = "Z"')
        cursor2.first()
        cursor2.setModeAccess(cursor2.Del)
        cursor2.refreshBuffer()
        cursor2.commitBuffer()

    def test_basic_3(self) -> None:
        """Basic tests 3."""

        from pineboolib.application.database import pnsqlcursor

        cursor = pnsqlcursor.PNSqlCursor("fltest", "default")

        self.assertEqual(cursor.msgCheckIntegrity(), "\nBuffer vacío o no hay metadatos")
        self.assertFalse(cursor.checkIntegrity(False))
        cursor.setModeAccess(cursor.Insert)
        cursor.refreshBuffer()
        self.assertEqual(cursor.msgCheckIntegrity(), "")
        self.assertTrue(cursor.checkIntegrity(False))

        self.assertFalse(cursor.cursorRelation())

        cursor3 = pnsqlcursor.PNSqlCursor("flareas")
        cursor3.setModeAccess(cursor.Insert)
        cursor3.refreshBuffer()
        cursor3.setValueBuffer("bloqueo", True)
        cursor3.setValueBuffer("idarea", "T")
        cursor3.setValueBuffer("descripcion", "Área de prueba T")
        self.assertTrue(cursor3.commitBuffer())

        cursor2 = pnsqlcursor.PNSqlCursor("flareas")
        cursor2.select()
        self.assertTrue(cursor2.first())
        cursor2.setUnLock("bloqueo", True)
        self.assertFalse(cursor2.isLocked())
        cursor2.setUnLock("bloqueo", False)
        self.assertFalse(cursor2.isLocked())
        cursor2.setModeAccess(cursor2.Del)
        cursor2.refreshBuffer()
        self.assertTrue(cursor2.commitBuffer())

        self.assertEqual(cursor2.curFilter(), "")
        cursor2.setFilter("bloqueo = true")
        self.assertEqual(cursor2.curFilter(), "bloqueo = true")

        cursor2.setSort("bloqueo ASC")

    def test_basic_4(self) -> None:
        """Basic tests 4."""
        from pineboolib.application.database import pnsqlcursor

        cursor_fake = pnsqlcursor.PNSqlCursor("fake")
        self.assertFalse(cursor_fake._valid)
        self.assertFalse(cursor_fake.isModifiedBuffer())
        self.assertTrue(cursor_fake.activatedCheckIntegrity())
        self.assertEqual(cursor_fake.checkIntegrity(False), False)
        self.assertEqual(cursor_fake.table(), "")
        self.assertEqual(cursor_fake.mainFilter(), "")
        cursor_fake.setValueBuffer("no_field", "XXX")
        self.assertFalse(cursor_fake.valueBuffer("no_field"))
        cursor = pnsqlcursor.PNSqlCursor("fltest")
        self.assertTrue(cursor._valid)
        cursor_2 = pnsqlcursor.PNSqlCursor("flareas")
        cursor_2.select()
        cursor_2.first()
        cursor_2.setModeAccess(cursor_2.Insert)
        self.assertFalse(cursor_2.fieldDisabled("descripcion"))
        cursor_2.refreshBuffer()
        self.assertEqual(
            cursor_2.msgCheckIntegrity(),
            "\nflareas:Área : No puede ser nulo\nflareas:Descripción : No puede ser nulo",
        )
        self.assertEqual(
            cursor_2.valueBuffer("descripcion"), cursor_2.valueBufferCopy("descripcion")
        )

        self.assertTrue(cursor_2.metadata().field("idarea"))
        buffer = cursor_2.buffer()
        if buffer:
            self.assertFalse(buffer.is_generated("idarea"))
        cursor_2.setEdition(True, "prueba")
        self.assertTrue(cursor_2.private_cursor.edition_states_)
        cursor_2.restoreEditionFlag("prueba")
        cursor_2.setBrowse(True, "prueba2")
        self.assertTrue(cursor_2.private_cursor.browse_states_)
        cursor_2.restoreBrowseFlag("prueba2")

    def test_basic_5(self) -> None:
        """Basic tests 5."""
        from pineboolib.application.database import pnsqlcursor
        from pineboolib import application
        from pineboolib.plugins.mainform.eneboo import eneboo
        from pineboolib.qsa import qsa

        application.PROJECT.main_window = eneboo.MainForm()
        application.PROJECT.main_window.initScript()

        cursor_6 = pnsqlcursor.PNSqlCursor("flareas")
        cursor_6.setModeAccess(cursor_6.Insert)
        cursor_6.refreshBuffer()
        cursor_6.setValueBuffer("bloqueo", True)
        cursor_6.setValueBuffer("idarea", "T")
        cursor_6.setValueBuffer("descripcion", "Área de prueba T")
        self.assertTrue(cursor_6.commitBuffer())

        cursor_qry = pnsqlcursor.PNSqlCursor("fltest2")
        self.assertTrue(cursor_qry)
        self.assertFalse(cursor_qry.private_cursor.needUpdate())
        cursor = pnsqlcursor.PNSqlCursor("fltest")
        self.assertFalse(cursor.private_cursor.needUpdate())
        self.assertEqual(cursor.primaryKey(), "id")
        self.assertEqual(cursor.fieldType("id"), 100)
        cursor_qry.setNotGenerateds()
        cursor.sort()

        self.assertFalse(cursor.fieldType("id2"))
        for field in cursor:
            self.assertTrue(field)

        cursor_2 = pnsqlcursor.PNSqlCursor("flareas")
        cursor_2.update(False)

        cursor.changeConnection("dbAux")
        self.assertEqual(cursor.transactionsOpened(), [])
        cursor.db().doTransaction(cursor)
        self.assertEqual(cursor.transactionsOpened(), ["1"])
        cursor.db().doRollback(cursor)
        cursor.setForwardOnly(True)

        cursor_3 = pnsqlcursor.PNSqlCursor("flareas")
        cursor_3.select()
        self.assertTrue(cursor_3.first())
        cursor_3.chooseRecord(False)
        qsa.from_project("formRecordflareas").reject()
        cursor_3.select()
        cursor_3.first()

        cursor_3.browseRecord(False)
        qsa.from_project("formRecordflareas").reject()

        cursor_3.select()
        cursor_3.first()

        cursor_3.editRecord(False)

        qsa.from_project("formRecordflareas").reject()

        cursor_4 = pnsqlcursor.PNSqlCursor("flareas")
        cursor_4.select()
        self.assertEqual(cursor_4.size(), 1)
        cursor_4.first()
        self.assertTrue(cursor_4.selection_pk("T"))
        cursor_4.last()
        self.assertTrue(cursor_4.selection_pk("T"))
        cursor_4.last()
        self.assertFalse(cursor_4.selection_pk("J"))
        cursor_4.setNull("idarea")
        cursor_4.setModeAccess(cursor_4.Insert)
        cursor_4.refreshBuffer()
        self.assertFalse(cursor_4.checkIntegrity())

    @classmethod
    def tearDown(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()


class TestRelations(unittest.TestCase):
    """Test Relations class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_basic_relations_1(self) -> None:
        """Test basic relations 1."""
        from pineboolib.application.database import pnsqlcursor

        cur_areas = pnsqlcursor.PNSqlCursor("flareas")

        cursor_6 = pnsqlcursor.PNSqlCursor("flareas")
        cursor_6.setModeAccess(cursor_6.Insert)
        cursor_6.refreshBuffer()
        cursor_6.setValueBuffer("bloqueo", True)
        cursor_6.setValueBuffer("idarea", "O")
        cursor_6.setValueBuffer("descripcion", "Área de prueba O")
        self.assertTrue(cursor_6.commitBuffer())
        rel = pnrelationmetadata.PNRelationMetaData(
            "flareas", "idarea", pnrelationmetadata.PNRelationMetaData.RELATION_1M
        )
        rel.setField("idarea")
        cur_areas.select()
        cur_areas.first()
        self.assertEqual(cur_areas.valueBuffer("idarea"), "O")
        cur_areas.setModeAccess(cur_areas.Edit)
        cur_areas.refreshBuffer()
        cur_modulos = pnsqlcursor.PNSqlCursor("flmodules", True, "default", cur_areas, rel)
        cur_modulos.select()
        cur_modulos.refreshBuffer()
        cur_rel = cur_modulos.cursorRelation()
        if cur_rel:
            self.assertEqual(cur_rel.valueBuffer("idarea"), "O")
        self.assertFalse(cur_areas.isLocked())
        self.assertFalse(cur_modulos.fieldDisabled("icono"))
        self.assertTrue(cur_modulos.isLocked())
        cur_modulos.setNull("icono")
        cur_modulos.setCopyNull("icono")
        self.assertTrue(cur_modulos.isCopyNull("icono"))
        cur_modulos.setModeAccess(cur_modulos.Insert)
        cur_modulos.refreshBuffer()
        cur_modulos.setValueBuffer("idmodulo", "TM")
        cur_modulos.setValueBuffer("descripcion", "Desc")
        self.assertEqual(cur_modulos.checkIntegrity(False), True)
        self.assertTrue(cur_modulos.commitBuffer())
        self.assertTrue(cur_areas.rollback())

    def test_basic_relations_2(self) -> None:
        """Test basic relations 2."""
        from pineboolib.application.database import pnsqlcursor
        from pineboolib.application.metadata import pnrelationmetadata

        cur_areas = pnsqlcursor.PNSqlCursor("flareas")
        rel = pnrelationmetadata.PNRelationMetaData(
            "flareas", "idarea", pnrelationmetadata.PNRelationMetaData.RELATION_1M
        )
        rel.setField("idarea")
        cur_areas.select()
        cur_areas.first()
        self.assertEqual(cur_areas.valueBuffer("idarea"), "O")
        cur_areas.setModeAccess(cur_areas.Edit)

        cur_areas.refreshBuffer()
        cur_areas.transaction()
        cur_modulos = pnsqlcursor.PNSqlCursor("flmodules", True, "default", cur_areas, rel)

        cur_modulos.setModeAccess(cur_modulos.Insert)
        self.assertEqual(cur_areas.transactionsOpened(), ["1"])

        cur_modulos.refreshBuffer()

        cur_modulos.rollbackOpened(-1, "Mensage de prueba 1º")

        cur_areas.setModeAccess(cur_areas.Edit)
        cur_areas.refreshBuffer()
        cur_areas.transaction()
        cur_modulos = pnsqlcursor.PNSqlCursor("flmodules", True, "default", cur_areas, rel)
        cur_modulos.setModeAccess(cur_modulos.Insert)
        self.assertEqual(cur_areas.transactionsOpened(), ["1", "2"])
        cur_modulos.refreshBuffer()
        cur_modulos.transaction()
        cur_modulos.commitOpened(-1, "Mensage de prueba 2º")

        cur_areas.setModeAccess(cur_areas.Edit)
        cur_areas.refreshBuffer()
        cur_areas.transaction()
        cur_modulos = pnsqlcursor.PNSqlCursor("flmodules", True, "default", cur_areas, rel)
        cur_modulos.setModeAccess(cur_modulos.Insert)
        self.assertEqual(cur_areas.transactionsOpened(), ["1", "2", "3"])
        cur_modulos.refreshBuffer()
        cur_modulos.transaction()
        self.assertTrue(cur_modulos.commitBufferCursorRelation())

        cur_areas.setModeAccess(cur_areas.Edit)
        cur_areas.refreshBuffer()
        cur_areas.transaction()
        self.assertTrue(cur_areas.commitBuffer())
        cur_areas.setEditMode()
        cur_areas.setModeAccess(cur_areas.Edit)
        cur_areas.setEditMode()
        cur_areas.prime_insert()
        cur_modulos.setModeAccess(cur_modulos.Insert)
        cur_modulos.refreshDelayed()
        self.assertTrue(cur_modulos.fieldDisabled("idarea"))
        self.assertFalse(cur_modulos.commitBuffer())
        self.assertEqual(cur_modulos.transactionsOpened(), ["4"])
        self.assertTrue(cur_modulos.rollback())
        self.assertEqual(cur_modulos.transactionsOpened(), [])
        self.assertTrue(cur_areas.rollback())
        self.assertEqual(cur_areas.transactionsOpened(), ["1", "2", "3"])
        cur_areas.setModeAccess(cur_areas.Insert)
        cur_areas.refreshBuffer()
        cur_areas.setValueBuffer("bloqueo", True)
        cur_areas.setValueBuffer("idarea", "Q")
        cur_areas.setValueBuffer("descripcion", "Área de prueba Q")

        cur_modulos.setModeAccess(cur_modulos.Insert)
        cur_modulos.refreshBuffer()
        cur_modulos.setValueBuffer("idmodulo", "CCC")
        cur_modulos.setValueBuffer("idarea", "Q")
        cur_modulos.setValueBuffer("descripcion", "modulo de prueba")
        cur_modulos.setValueBuffer("version", "0.0")

        self.assertTrue(cur_modulos.commitBufferCursorRelation())
        cur_modulos.setModeAccess(cur_modulos.Edit)
        self.assertTrue(cur_modulos.commitBufferCursorRelation())

    def test_basic_relations_3(self) -> None:
        """Test basic relations 3."""

        cursor = pnsqlcursor.PNSqlCursor("flmodules")
        cursor.select()
        cursor.setModeAccess(cursor.Insert)
        cursor.setValueBuffer("idarea", "")
        cursor.setValueBuffer("idmodulo", "WWW")
        cursor.setValueBuffer("descripcion", "TEST")
        cursor.setValueBuffer("version", "0.0")
        self.assertFalse(cursor.commitBuffer())

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()


class TestAcos(unittest.TestCase):
    """Test Acos class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def testSetAcosCondition(self) -> None:
        """Test setAcosCondition."""

        from pineboolib.application.database import pnsqlcursor

        cur_grupos = pnsqlcursor.PNSqlCursor("flgroups")
        cur_grupos.setModeAccess(cur_grupos.Insert)
        cur_grupos.refreshBuffer()
        cur_grupos.setValueBuffer("idgroup", "a")
        cur_grupos.setValueBuffer("descripcion", "desc a")
        self.assertTrue(cur_grupos.commitBuffer())
        cur_grupos.setModeAccess(cur_grupos.Insert)
        cur_grupos.refreshBuffer()
        cur_grupos.setValueBuffer("idgroup", "b")
        cur_grupos.setValueBuffer("descripcion", "desc b")
        self.assertTrue(cur_grupos.commitBuffer())
        cur_grupos.setModeAccess(cur_grupos.Insert)
        cur_grupos.refreshBuffer()
        cur_grupos.setValueBuffer("idgroup", "c")
        cur_grupos.setValueBuffer("descripcion", "desc c")
        self.assertTrue(cur_grupos.commitBuffer())
        cur_grupos.setModeAccess(cur_grupos.Insert)
        cur_grupos.refreshBuffer()
        cur_grupos.setValueBuffer("idgroup", "d")
        cur_grupos.setValueBuffer("descripcion", "desc d")
        self.assertTrue(cur_grupos.commitBuffer())

        cur_grupos.select()
        field = cur_grupos.metadata().field("descripcion")
        self.assertTrue(field is not None)
        if field is not None:
            while cur_grupos.next():
                self.assertTrue(field.editable())

        cur_grupos.setAcTable("r-")
        cur_grupos.setAcosCondition("descripcion", cur_grupos.Value, "desc c")

        cur_grupos.select()
        while cur_grupos.next():
            field_2 = cur_grupos.metadata().field("descripcion")

            self.assertTrue(field_2 is not None)
            if field_2 is not None:
                if cur_grupos.valueBuffer("descripcion") == "desc c":
                    self.assertFalse(field_2.editable())
                else:
                    self.assertTrue(field_2.editable())

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()


class TestAfterCommit(unittest.TestCase):
    """Test AfterCommit class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        from pineboolib.qsa import qsa

        application.VIRTUAL_DB = True
        init_testing()
        session = qsa.session()
        session.rollback()
        session.connection().close()

    def test_basic_1(self) -> None:
        """Test sys.afertCommit_flfiles is called"""

        from pineboolib.plugins.mainform.eneboo import eneboo
        from pineboolib.qsa import qsa

        application.PROJECT.main_window = eneboo.MainForm()
        application.PROJECT.main_window.initScript()

        util = qsa.FLUtil()

        self.assertEqual(util.sqlSelect("flserial", "sha", "1=1"), False)
        qsa_sys = qsa.sys
        path = fixture_path("principal.eneboopkg")
        self.assertTrue(qsa_sys.loadModules(path, False))
        qsa.from_project("flfactppal").iface.valoresIniciales()
        qsa.from_project("sys").iface.init()

    def test_basic_2(self) -> None:
        """Test size and sha."""
        from pineboolib.qsa import qsa

        util = qsa.FLUtil()
        size_ = util.sqlSelect("flfiles", "COUNT(*)")
        self.assertTrue(size_ in [146, 147])
        sha_ = util.sqlSelect("flserial", "sha", "1=1")
        if size_ == 147:
            self.assertEqual(sha_, "79D7F8BEFE9C4ECAA33E3D746A86586EFC90AB86")

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        from pineboolib.qsa import qsa

        session = qsa.session()
        session.rollback()

        finish_testing()


class TestCorruption(unittest.TestCase):
    """Test Acos class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""

        # application.VIRTUAL_DB = True
        init_testing()

    def test_basic_1(self) -> None:
        """Test Populate cursor."""

        from pineboolib.application.database import pnsqlcursor

        cursor = pnsqlcursor.PNSqlCursor("fltest")

        for i in range(100):
            cursor.setModeAccess(cursor.Insert)
            cursor.refreshBuffer()
            cursor.setValueBuffer("string_field", "Linea %s" % i)
            print("*", i)
            self.assertTrue(cursor.commitBuffer())

    def test_basic_2(self) -> None:
        """Test Deleteting."""

        from pineboolib.application.database import pnsqlcursor
        from pineboolib.qsa import qsa

        cursor = pnsqlcursor.PNSqlCursor("fltest")
        cursor.select()
        self.assertEqual(cursor.size(), 100)

        cursor.select("string_field ='Linea 10'")
        self.assertEqual(cursor.size(), 1)
        cursor.select()
        self.assertEqual(cursor.size(), 100)
        self.assertTrue(cursor.first())
        cursor.setModeAccess(cursor.Edit)
        self.assertTrue(cursor.refreshBuffer())
        cursor.setValueBuffer("string_field", "Linea 10 m.")

        self.assertTrue(qsa.FLUtil().sqlDelete("fltest", "string_field='Linea 10'"))

        cursor_2 = pnsqlcursor.PNSqlCursor("fltest")
        cursor_2.select()

        self.assertEqual(cursor_2.size(), 99)

        self.assertTrue(cursor.commitBuffer())

        cursor.select("string_field ='Linea 9'")
        self.assertEqual(cursor.size(), 1)
        self.assertTrue(cursor.first())
        cursor.setModeAccess(cursor.Edit)
        self.assertTrue(cursor.refreshBuffer())
        cursor.setValueBuffer("string_field", "Linea 9 m.")
        self.assertTrue(cursor.commitBuffer())
        cursor.refresh()
        self.assertEqual(cursor.size(), 99)

        cursor_3 = pnsqlcursor.PNSqlCursor("fltest")
        cursor_3.setSort("string_field ASC")
        cursor_3.select()

        cursor_3.setForwardOnly(True)
        self.assertEqual(cursor_3.size(), 99)
        i = 1
        while cursor_3.next():
            if i == 10:
                self.assertTrue(qsa.FLUtil().sqlDelete("fltest", "string_field = 'Linea 20'"))
                i += 1

            self.assertEqual(cursor_3.size(), 99)
            i += 1

        cursor_3.setForwardOnly(False)
        cursor_3.select()
        self.assertEqual(cursor_3.size(), 98)

    def test_out_transaction(self) -> None:
        """Out transaction test."""

        from pineboolib.application.database import pnsqlcursor
        from pineboolib.fllegacy import flutil

        cursor = pnsqlcursor.PNSqlCursor("Fltest3")
        cursor.setModeAccess(cursor.Insert)
        cursor.refreshBuffer()
        cursor.setValueBuffer("bool_field", False)
        pk_value = cursor.valueBuffer(cursor.primaryKey())
        self.assertTrue(cursor.commitBuffer())
        cursor.select("counter_field='%s'" % pk_value)
        self.assertTrue(cursor.first())
        cursor.setModeAccess(cursor.Edit)
        cursor.refreshBuffer()
        valor_inicial = cursor.valueBuffer("bool_field")
        valor_inicial_sql = flutil.FLUtil.sqlSelect(
            "fltest3", "bool_field", "counter_field='%s'" % pk_value
        )
        self.assertEqual(valor_inicial, valor_inicial_sql)
        cursor.setValueBuffer("bool_field", True)
        valor_intermedio = cursor.valueBuffer("bool_field")
        self.assertNotEqual(
            valor_inicial,
            valor_intermedio,
            "la pk era %s" % cursor.valueBuffer(cursor.primaryKey()),
        )
        valor_intermedio_sql = flutil.FLUtil.sqlSelect(
            "fltest3", "bool_field", "counter_field='%s'" % pk_value
        )
        self.assertEqual(valor_intermedio, valor_intermedio_sql)
        cursor.setValueBuffer("bool_field", False)
        valor_final = cursor.valueBuffer("bool_field")
        self.assertNotEqual(valor_final, valor_intermedio)
        valor_final_sql = flutil.FLUtil.sqlSelect(
            "fltest3", "bool_field", "counter_field='%s'" % pk_value
        )
        self.assertEqual(valor_final, valor_final_sql)
        self.assertTrue(cursor.metadata().field("bool_field").outTransaction())  # type: ignore [union-attr]

    def test_basic_3(self) -> None:
        """Bad cursor."""

        from pineboolib.application.database import pnsqlcursor
        from pineboolib.fllegacy import flutil

        cursor = pnsqlcursor.PNSqlCursor("Fltest")
        self.assertEqual(cursor.metadata().name(), "fltest")
        cursor.setAction("bad_Action")
        action = cursor.action()
        if action:
            self.assertEqual(action.name(), "bad_action")
        cursor.select()
        self.assertEqual(cursor.size(), 0)
        cursor.setAction("fltest")

        cursor.select()
        self.assertEqual(cursor.size(), 98)
        util = flutil.FLUtil()

        util.sqlDelete("fltest", "1=1", "dbAux")
        cursor.select()
        self.assertEqual(cursor.size(), 0)

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()



"""Flloadmod module."""
# -*- coding: utf-8 -*-
from pineboolib.qsa import qsa


class FormInternalObj(qsa.FormDBWidget):
    """FormInternalObj class."""

    def main(self) -> None:
        """Entry function."""
        continuar = qsa.MessageBox.warning(
            qsa.util.translate(
                "scripts",
                "Antes de cargar un módulo asegúrese de tener una copia de seguridad de todos los datos,\n"
                + "y de que no hay ningun otro usuario conectado a la base de datos mientras se realiza la carga.\n\n¿Desea continuar?",
            ),
            qsa.MessageBox.Yes,
            qsa.MessageBox.No,
        )
        if continuar == qsa.MessageBox.No:
            return
        nombre_fichero = qsa.FileDialog.getOpenFileName(
            "modfiles(*.mod)", qsa.util.translate("scripts", "Elegir Fichero")
        )
        if nombre_fichero:
            fichero = qsa.File(nombre_fichero)
            if not qsa.from_project("formRecordflmodules").aceptarLicenciaDelModulo(
                qsa.ustr(fichero.path, "/")
            ):
                qsa.MessageBox.critical(
                    qsa.util.translate(
                        "scripts", "Imposible cargar el módulo.\nLicencia del módulo no aceptada."
                    ),
                    qsa.MessageBox.Ok,
                )
                return

            if qsa.from_project("formflreloadlast").load_module(nombre_fichero):
                qsa.aqApp.reinit()

    def get_value(self, linea: str) -> str:
        """Return value."""
        return linea


def version_compare(ver1: str, ver2: str) -> int:
    """Compare two versions and return the hightest."""
    return qsa.from_project("formflreloadlast").version_compare(ver1, ver2)


def deps_evaluate(deps: qsa.Array) -> bool:
    """Evaluate dependencies."""

    for dep in deps:
        if not qsa.sys.isLoadedModule(dep):
            res = qsa.MessageBox.warning(
                qsa.util.translate("scripts", "Este módulo depende del módulo ")
                + dep
                + qsa.util.translate(
                    "scripts",
                    ", que no está instalado.\nFacturaLUX puede fallar por esta causa.\n¿Desea continuar la carga?",
                ),
                qsa.MessageBox.Yes,
                qsa.MessageBox.No,
            )
            if res == qsa.MessageBox.No:
                return False

    return True



"""Test for basemodel module."""

import unittest

from pineboolib.loader.main import init_testing, finish_testing
from pineboolib.qsa import qsa


class TestBaseModel(unittest.TestCase):
    """TestBaseModel Class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_default_value(self) -> None:
        """Test default values when new instances."""

        obj_ = qsa.orm_("flareas")()
        self.assertEqual(obj_.bloqueo, True)

        self.assertTrue(obj_.get_session_from_connection())
        self.assertTrue(obj_.copy())

    def test_cursor_mode_access(self) -> None:
        """Test mode access."""

        obj_ = qsa.orm.flareas()
        self.assertTrue(obj_.module_iface)

        obj2_ = qsa.orm.flareas()
        self.assertTrue(obj2_.module_iface)

        cursor = obj_.cursor
        self.assertEqual(cursor.modeAccess(), cursor.Insert)
        obj_.idarea = "TT"
        self.assertEqual(cursor.modeAccess(), cursor.Insert)

        obj1_ = qsa.orm.flareas.get("O")
        self.assertTrue(obj1_.module_iface)
        self.assertFalse(obj1_.changes())
        self.assertEqual(obj1_.pk, "O")
        self.assertEqual(obj1_.cursor.modeAccess(), obj1_.cursor.Browse)
        obj1_.descripcion = "prueba nuevo"
        obj1_.pk = "O"
        self.assertTrue(obj1_.changes())
        self.assertEqual(obj1_.cursor.modeAccess(), obj1_.cursor.Edit)
        obj1_.allow_buffer_changed("idarea", False)
        self.assertTrue("idarea" in obj1_._deny_buffer_changed)
        obj1_.allow_buffer_changed("idarea", True)
        self.assertFalse("idarea" in obj1_._deny_buffer_changed)

    def test_2_metadata(self) -> None:
        """Test table_metadata."""
        session = qsa.session()
        obj_class = qsa.orm_("fltest")
        obj_ = obj_class(session=session)
        meta = obj_.table_metadata()
        self.assertTrue(meta)
        session.begin()
        self.assertTrue(obj_.save())
        obj_.session.commit()

    def test_serial(self) -> None:
        """Test serial field."""
        qsa.thread_session_new()
        class_fltest = qsa.orm_("fltest")
        obj_ = class_fltest()
        self.assertEqual(obj_.id, 2)

        obj2_ = class_fltest(serial=False)
        self.assertEqual(obj2_.id, None)

        class_fltest2 = qsa.orm_("fltest")
        obj3_ = class_fltest2()
        self.assertEqual(obj3_.id, 3)

    def test_3_get(self) -> None:
        """Test get classmethod."""

        class_fltest = qsa.orm_("fltest")
        session = qsa.session()
        obj_ = class_fltest.get(1, session)
        self.assertTrue(obj_)

    #
    #         self.assertTrue(obj_)
    #         obj_2 = class_fltest.get(1)
    #         self.assertTrue(obj_2)
    #         self.assertEqual(obj_, obj_2)
    #         obj_3 = class_fltest.query().get(1)
    #         self.assertEqual(obj_, obj_3)
    #         self.assertTrue(obj_3)
    #         obj_4 = class_fltest.query().get(2)
    #         self.assertFalse(obj_4)
    #         self.assertEqual(obj_3.id, 1)
    #         self.assertNotEqual(obj_, obj_4)
    #
    #         obj_5 = class_fltest.get(1)
    #         self.assertTrue(obj_5.delete())
    #         obj_5.session.commit()
    # ===============================================================================

    def test_integrity(self) -> None:
        """test _check_integrity."""

        session = qsa.thread_session_current()
        qsa.thread_session_free()
        new_session = qsa.thread_session_new()
        self.assertNotEqual(session, new_session)
        obj_ = qsa.orm_("flmodules")()
        obj_.idmodulo = "mod2"
        obj_.idarea = "F"
        with self.assertRaises(Exception):
            obj_.save()

        obj_.descripcion = "PRUEBA"
        with self.assertRaises(Exception):
            obj_.save()

        obj_2 = qsa.orm_("flareas")()
        obj_2.idarea = "F"
        obj_2.descripcion = "Area"
        self.assertTrue(obj_2.save())

        self.assertTrue(obj_.relationM1("idarea"))
        self.assertEqual(obj_.get_transaction_level(), -1)

        self.assertTrue(obj_.save())
        self.assertEqual(qsa.FLUtil().sqlSelect("flmodules", "idmodulo", "idarea='F'"), "mod2")
        obj_3 = qsa.orm_("flmodules")()
        obj_3.idmodulo = "mod1"
        obj_3.idarea = "G"
        obj_3.descripcion = "PRUEBA"

        self.assertEqual(
            obj_3.changes(),
            {
                "bloqueo": True,
                "descripcion": "PRUEBA",
                "idarea": "G",
                "idmodulo": "mod1",
                "version": "0.0",
            },
        )

        self.assertFalse(obj_3.is_being_changed())
        self.assertTrue(obj_3.is_being_created())
        self.assertFalse(obj_3.is_being_deleted())

        obj_3.mode_access = 3
        self.assertEqual(obj_3.mode_access, 3)

    def test_integrity_2(self) -> None:
        """Test integrity"""

        orm = qsa.orm

        class_modulos = qsa.orm.flmodules
        self.assertTrue(class_modulos)
        mod_1 = class_modulos()
        mod_1.idmodulo = "prueba"
        mod_1.descripcion = "descripcion"
        mod_1.idarea = ""
        with self.assertRaises(Exception):
            mod_1._check_integrity()

        class_test4 = orm.fltest4
        self.assertTrue(class_test4)
        obj_ = class_test4()
        obj_.idarea = 22
        obj_.id_test = 0
        obj_.other_field = "NO"
        # with self.assertRaises(Exception):
        #    obj_._check_integrity()

        obj_.id_test = 1
        obj_._check_integrity()

    def test_relation_m1(self) -> None:
        """Test relationM1."""
        qsa.thread_session_free()
        qsa.session()
        current_session = qsa.thread_session_current()
        if current_session is not None:
            current_session.begin()
            obj_ = qsa.orm_("flareas")()
            obj_.idarea = "T"
            obj_.descripcion = "Area"
            self.assertTrue(obj_.save())
            obj_.session.commit()

            obj_2 = qsa.orm_("flmodules")()
            obj_2.idmodulo = "mod1"
            obj_2.idarea = "T"
            obj_2.descripcion = "PRUEBA relation M1"

            obj_rel = obj_2.relationM1("idarea")
            obj_rel_1 = obj_2.relationM1("idmodulo")
            self.assertFalse(obj_rel_1)
            self.assertTrue(obj_rel)
            self.assertEqual(obj_rel.idarea, obj_.idarea)

    def test_relation_1m(self) -> None:
        """Test realtion 1M."""

        obj_class = qsa.orm_("flareas")

        obj_2 = qsa.orm_("flmodules")()
        obj_2.idmodulo = "mod4"
        obj_2.idarea = "F"
        obj_2.descripcion = "relation_m1"
        self.assertTrue(obj_2.save())
        obj_ = obj_class.query(obj_2.session).get("F")
        self.assertTrue(obj_)
        # obj_2.session.commit()
        relations_dict = obj_.relation1M("idarea")
        modules_rel = relations_dict["flmodules_idarea"]
        self.assertTrue(obj_2)
        self.assertEqual(len(modules_rel), 2)
        self.assertEqual(modules_rel[1].idmodulo, obj_2.idmodulo, modules_rel)

    def test_base(self) -> None:
        """Test."""
        session = qsa.session()
        session.begin()
        obj_class = qsa.orm_("flareas")
        obj_ = obj_class(session=session)
        self.assertEqual(obj_.mode_access, 0)  # Insert
        obj_.idarea = "O"
        obj_.descripcion = "Descripcion O"
        self.assertTrue(obj_.save())
        obj_.session.commit()

        session.begin()
        obj_new = obj_class.get("O", session)
        obj_new.descripcion = "Nueva descripción"
        self.assertTrue(obj_new.is_being_changed())
        self.assertFalse(obj_new.is_being_created())
        self.assertTrue(obj_new.save())
        session.commit()

    def test_cache_objects(self) -> None:
        """Test cache objects."""
        qsa.thread_session_new()
        obj_class = qsa.orm_("flareas")
        obj_ = obj_class()
        obj_.idarea = "R"
        obj_.descripcion = "Descripción de R"
        self.assertTrue(obj_.save())

        obj_2 = obj_class.query(obj_.session).all()[1]
        self.assertTrue(obj_2)
        self.assertEqual(obj_, obj_2, obj_class.query(obj_.session).all())
        obj_2.descripcion = "Descripción de P"
        self.assertEqual(obj_.descripcion, "Descripción de P")

    def test_counter(self) -> None:
        """Test counter."""
        qsa.thread_session_new()
        obj_class = qsa.orm_("fltest3")
        obj_ = obj_class(counter=True)
        self.assertEqual(obj_.counter_field, "000001")
        obj_class = qsa.orm_("fltest3")
        obj2_ = obj_class()
        self.assertEqual(obj2_.counter_field, None)

    def test_z_delete(self) -> None:
        """Test delete."""

        qsa.thread_session_free()
        self.assertFalse(qsa.thread_session_current())

        session = qsa.thread_session_new()
        obj_class = qsa.orm.flareas

        obj_ = obj_class.get("F")
        self.assertEqual(session, obj_.session)

        self.assertTrue(obj_)
        self.assertEqual(obj_class.query().all()[2].idarea, obj_.idarea)

        self.assertFalse(obj_.relation1M())

        obj_2_class = qsa.orm_("flmodules")
        obj2_ = obj_2_class()
        obj2_.idarea = "F"
        obj2_.descripcion = "Desc"
        obj2_.idmodulo = "mr1"
        self.assertTrue(obj2_.save())
        self.assertEqual(session, obj2_.session)
        session.begin_nested() if session.transaction else session.begin()

        obj2_.session.commit()
        session.begin_nested() if session.transaction else session.begin()
        self.assertEqual(len(obj_.relation1M("idarea")["flmodules_idarea"]), 3)
        self.assertEqual(obj_.relation1M("idarea")["flmodules_idarea"][2].idmodulo, obj2_.idmodulo)
        self.assertTrue(obj_.delete())
        obj_.session.commit()
        # self.assertFalse(obj_.relation1M("idarea")["flmodules_idarea"])

    def test_z_real_relation(self) -> None:
        """Test real relation."""

        areas_class = qsa.orm_("flareas")
        modules_class = qsa.orm_("flmodules")

        current_session = qsa.thread_session_current()
        if current_session:
            if not current_session.transaction:
                current_session.begin()

            obj_areas = areas_class()
            obj_areas.idarea = "I"
            obj_areas.descripcion = "Descripción I"
            self.assertTrue(obj_areas.save())
            obj_areas.session.commit()

            obj_modules_1 = modules_class()
            obj_modules_1.idarea = "I"
            obj_modules_1.descripcion = "modulo 1"
            obj_modules_1.idmodulo = "M1"

            obj_modules_2 = modules_class()
            obj_modules_2.idarea = "I"
            obj_modules_2.descripcion = "modulo 2"
            obj_modules_2.idmodulo = "M2"

            current_session.begin()

            self.assertTrue(obj_modules_1.save())
            self.assertTrue(obj_modules_2.save())

            obj_modules_1.session.commit()

            # self.assertEqual(obj_modules_1.parent[0], obj_areas)
            # self.assertEqual(obj_modules_2.parent[0], obj_areas)
            self.assertEqual(
                len(modules_class.query().filter(modules_class.idarea == "I").all()), 2
            )
            self.assertEqual(len(obj_areas.children), 2)

            # for child in obj_areas.children: #Modo correcto para lanzar eventos ... si no hay legacy_metadata.deleteCascade()
            #    self.assertTrue(child.delete())
            obj_areas.session.begin()
            self.assertTrue(obj_areas.delete())
            obj_areas.session.commit()
            self.assertEqual(
                len(modules_class.query().filter(modules_class.idarea == "I").all()), 0
            )

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""

        finish_testing()



"""
ModuleActions module.
"""

from pineboolib.core import exceptions
from pineboolib.core.utils import utils_base
from pineboolib.application import xmlaction
from pineboolib import application, logging

from typing import Any, TYPE_CHECKING, NoReturn

if TYPE_CHECKING:
    from pineboolib.application import module
    from pineboolib.application import projectmodule

LOGGER = logging.get_logger(__name__)


class ModuleActions(object):
    """
    Generate tree with actions from modules.
    """

    module_name: str
    path: str
    mod: "module.Module"
    project: "projectmodule.Project"

    def __init__(self, module_: "module.Module", path: str, modulename: str) -> None:
        """
        Initialize.

        @param module. Identificador del módulo
        @param path. Ruta del módulo
        @param modulename. Nombre del módulo
        """

        self.project = (
            module_  # type: ignore [assignment] # noqa: F821
            if TYPE_CHECKING
            else application.PROJECT
        )

        self.mod = module_
        self.path = path
        self.module_name = modulename
        if not self.path:
            LOGGER.error("El módulo no tiene un path válido %s", self.module_name)

    def load(self) -> None:
        """Load module actions into project."""
        # Ojo: Almacena un arbol con los módulos cargados
        from pineboolib.application.qsadictmodules import QSADictModules

        tree = utils_base.load2xml(self.path)
        self.root = tree.getroot()

        action = xmlaction.XMLAction(self, self.mod.name)
        if action is None:
            raise Exception("action is empty!")

        # action._mod = self
        # action.alias = self.mod.name
        # action.form = self.mod.name
        # action._form = None
        # action.table = None
        # action.scriptform = self.mod.name
        self.project.actions[
            action._name
        ] = action  # FIXME: Actions should be loaded to their parent, not the singleton
        QSADictModules.save_action_for_root_module(action)

        for xmlaction_item in self.root:
            action_xml = xmlaction.XMLAction(self, xmlaction_item)
            name = action_xml._name
            if name in ("unnamed", ""):
                continue

            if (
                QSADictModules.save_action_for_mainform(action_xml)
                or name not in self.project.actions.keys()
            ):
                self.project.actions[
                    name
                ] = action_xml  # FIXME: Actions should be loaded to their parent, not the singleton
            QSADictModules.save_action_for_formrecord(action_xml)
            QSADictModules.save_action_for_class(action_xml)

        if [
            file_.name
            for file_ in self.project.files.values()
            if file_.name.startswith("plus_sys") and file_.module == self.module_name
        ]:
            action_xml_plus = xmlaction.XMLAction(self, "plus_sys")
            if QSADictModules.save_action_for_mainform(action_xml_plus):
                LOGGER.warning("plus_sys loaded!")
                self.project.actions["plus_sys"] = action_xml_plus

    def __contains__(self, name: str) -> bool:
        """Determine if it is the owner of an action."""
        return (
            name in self.project.actions
        )  # FIXME: Actions should be loaded to their parent, not the singleton

    def __getitem__(self, name: str) -> Any:
        """
        Retrieve particular action by name.

        @param name. Nombre de la action
        @return Retorna el XMLAction de la action dada
        """
        return self.project.actions[
            name
        ]  # FIXME: Actions should be loaded to their parent, not the singleton

    def __setitem__(self, name: str, action_: "xmlaction.XMLAction") -> "NoReturn":
        """
        Add action to a module property.

        @param name. Nombre de la action
        @param action_. Action a añadir a la propiedad del módulo
        """
        raise exceptions.ForbiddenError("Actions are not writable!")
        # self.project.actions[name] = action_
